
import sys, os, shutil, time
import traceback
import numpy as np
import torch
import yaml
from torch import multiprocessing as mp

import omnifig as fig

from bisect import bisect_left

from .. import util

# import tensorflow as tf
# from tensorboard import main as tb
from tensorboard import program
import pandas as pd

try:
	import matplotlib.pyplot as plt
	from matplotlib.figure import Figure
	from matplotlib.animation import Animation
	import seaborn as sns
except ImportError:
	print('WARNING: matplotlib not found')

try:
	from IPython.display import HTML
except ImportError:
	print('WARNING: ipython not found')

class Visualization(object): # wriapper
	def __init__(self, obj):
		self.obj = obj

	def view(self, scale=1):

		if isinstance(self.obj, util.Video):
			return HTML(self.obj.as_animation(scale=scale).to_html5_video())

		return self.obj

	def __str__(self):
		return 'Viz({})'.format(self.obj)

	def __repr__(self):
		return repr(self.obj)

	def save(self, path, vid_ext='mp4', img_ext='png', **kwargs):
		if isinstance(self.obj, Animation):
			if 'fps' not in kwargs:
				kwargs['fps'] = 30
			if 'writer' not in kwargs:
				kwargs['writer'] = 'imagemagick'

			if 'gif' not in path and 'mp4' not in path:
				path = '{}.{}'.format(path, vid_ext)

			self.obj.save(path, **kwargs)

		elif isinstance(self.obj, Figure):
			if '.pdf' not in path and '.png' not in path: # only png or pdf
				path = '{}.{}'.format(path, img_ext)
			self.obj.savefig(path)

		elif isinstance(self.obj, util.Video):

			if '.gif' not in path and '.mp4' not in path: # only gif or mp4
				path = '{}.{}'.format(path, vid_ext)

			self.obj.export(path)

		else:
			raise Exception('Unknown obj: {}'.format(type(self.obj)))


class Run_Manager(object):
	def __init__(self, names, root=None, override=None, pbar=None):
		
		self._names = names
		self._root = root
		self._override = override
		self._pbar = pbar
		
		self._load_runs()
		
	def reset(self):
		self._runs = []
		self.active = self._runs
		
	def _load_runs(self):
		self.reset()
		kwargs = {}
		if self._root is not None:
			kwargs['root'] = str(self._root)
		if self._override is not None:
			kwargs['override'] = self._override.copy()
		todo = [None]*len(self._names), self._names
		if isinstance(self._names, dict):
			todo = self._names.items()
		todo = todo if self._pbar is None else self._pbar(todo)
		for title, name in todo:
			load_config = fig.get_config(load=name, **kwargs)
			load_config.set_silent(True)
			run = fig.run('load-run', load_config)
			if title is not None:
				run.ident = title
			self._runs.append(run)
		self.active = self._runs.copy()

	def __len__(self):
		return len(self.active)

	def __iter__(self):
		return iter(self.active)

	def remove(self, name):

		if name not in self:
			raise ValueError(name)

		del self[self(name).idx]

	def __delitem__(self, item):
		del self.active[item]
		self.set_active(self.active)

	def map(self, fn, safe=False, pbar=None, reduce=None):
		'''
		fn is a callable taking one run as input
		'''

		if pbar is None:
			pbar = self._pbar

		outs = []

		seq = self.active if pbar is None else pbar(self.active)

		for run in seq:
			try:
				out = fn(run)
				outs.append(out)
			except Exception as e:
				if safe:
					print('{} failed'.format(run.get_name()))
					traceback.print_exc()
				else:
					raise e

		if pbar is not None:
			seq.close()

		if reduce is not None:
			return reduce(outs)
		return outs

	def through(self, **map_kwargs):

		def _execute(fn, args=[], kwargs={}):
			return self.map(lambda run: fn(run, *args, **kwargs),
			                **map_kwargs)

		return util.make_ghost(Run, _execute)

	def clear_run_cache(self, **kwargs):
		self.map(lambda run: run.purge(), **kwargs)

	def load_results(self, ident, load_kwargs={}, **kwargs):
		return self.map(lambda run: run.get_results(ident, **load_kwargs), **kwargs)

	# def stats_dataframe(self): # gets all stats
	#
	# 	data = []
	#
	# 	keys = None
	#
	# 	for run in self.active:
	# 		if 'stats' in run:
	# 			if keys is None:
	# 				keys = list(run.stats.keys())
	#
	# 			row = [run.stats[key] for key in keys]
	#
	# 			data.append(row)
	#
	# 	if keys is None:
	# 		print('No stats found')
	# 		return None
	#
	# 	return pd.DataFrame(data, columns=keys)



	def plot_stats(self, *stats, figax=None, get_X=None, palette='muted', **kwargs):

		df = self.stats_dataframe()[list(stats)]

		if get_X is None:
			get_X = lambda run: run.idx

		X = [get_X(run) for run in self]

		df['run'] = X

		yname = 'value'
		hue = 'stat'

		if len(stats) > 1:
			df = pd.melt(df, id_vars='run', var_name=hue, value_name=yname)
		else:
			yname = stats[0]
			hue = None
			palette = None

		if figax is None:
			figax = plt.subplots()
		_, ax = figax

		sns.barplot(x='run', y=yname, hue=hue, data=df, ax=ax, palette=palette, **kwargs)

		# plt.title(', '.join(stats))

		return figax


	# def prep_info(self, checkpoint=None, load_last=True, clear_info=False, force=False, name=None):
	#
	# 	self.select_checkpoint(checkpoint=checkpoint, load_last=load_last, force=force, name=name)
	# 	self.load_configs(force=force, clear_info=clear_info)

	# def select_checkpoint(self, checkpoint=None, load_last=True, force=False, name=None):
	#
	# 	note = ('last' if load_last else 'best') if checkpoint is None else checkpoint
	# 	print('Selecting checkpoint: {}'.format(note))
	#
	# 	valid = []
	#
	# 	for run in self.active:
	# 		if 'ckpt_path' not in run or force:
	#
	# 			if name is None:
	#
	# 				if checkpoint is None:
	# 					ckpt_path = find_checkpoint(run.path, load_last=load_last)
	# 				else:
	# 					ckpt_path = os.path.join(run.path, 'checkpoint_{}.pth.tar'.format(checkpoint))
	#
	# 				name_req = 'checkpoint_'
	# 			else:
	# 				name_req = name
	# 				ckpt_path = os.path.join(run.path, name)
	#
	# 			if name_req not in ckpt_path or not os.path.isfile(ckpt_path):
	# 				print('{} has no checkpoint'.format(run.name))
	# 				continue
	#
	# 			run.ckpt_path = ckpt_path
	#
	# 			ckpt_name = os.path.basename(run.ckpt_path)
	#
	# 			if name is None:
	# 				try:
	# 					ckpt_num = int(ckpt_name.split('_')[-1].split('.')[0])
	# 					run.meta.ckpt = ckpt_num
	# 				except ValueError as e:
	# 					print(run.ckpt_path, ckpt_name)
	# 					raise e
	#
	# 		valid.append(run)
	#
	# 	self.set_active(valid)
	#
	# def load_records(self, pbar=None, force=False):
	#
	# 	runs = self.active
	# 	if pbar is not None:
	# 		runs = pbar(runs)
	# 		runs.set_description('Loading records')
	#
	# 	for run in runs:
	# 		if 'records' not in run or force:
	# 			ckpt = torch.load(run.ckpt_path)
	# 			run.records = ckpt['records']
	#
	#
	# def load_configs(self, force=False, clear_info=False, excluded_info=[],
	#                  update_active=True, **kwargs):
	#
	# 	excluded_info = set(excluded_info)
	#
	# 	self.map(lambda r: r.load_config(force=force, clear_info=clear_info, excluded_info=excluded_info), **kwargs)
	#
	# 	if update_active:
	# 		valid = [run for run in self.active if 'config' in run]
	# 		self.set_active(valid)
	# 	print('Loaded configs')

	def set_active(self, active=None):
		if active is None:
			active = self.full_info.copy()
		self.active = active
		for i, run in enumerate(self.active):
			run.idx = i
		self.name2idx = None
		self.output = None
	def extend(self, runs):
		new = []
		for item in runs:
			if item.name not in self:
				new.append(item)
		if len(new):
			self.set_active(self.active + new)
	def append(self, item):
		if item in self: # dont allow duplicates
			return
		self.append(item)
		self.set_active(self.active)

	def add(self, *names):
		runs = []

		for name in names:
			for run in self.full_info:
				if name == run.name:
					runs.append(run)

		if len(runs):
			self.extend(runs)

			print('Added {} run(s)'.format(len(runs)))

	def clear(self):
		self.set_active([])

	def store_selection(self, name, runs=None):
		self.selections[name] = self.active.copy() if runs is None else runs

	def switch_selection(self, name):
		self.set_active(self.selections[name])

	def view_selections(self):
		for k,v in self.selections.items():
			print('{} : holds {} runs'.format(k, len(v)))

	def clear_selections(self):
		self.selections.clear()

	def invert(self):
		inv = [r for r in self.full_info if r.name not in self]
		self.set_active(inv)

	def _update_names(self):
		if self.name2idx is not None:
			return
		self.name2idx = {run.name:i for i,run in enumerate(self.active)}

	def __call__(self, name):
		if isinstance(name, int):
			return self.active[name]
		if isinstance(name, str):
			self._update_names()
			idx = self.name2idx[name]
			return self.active[idx]

		raise Exception('{} not found'.format(name))

	def __contains__(self, item):
		self._update_names()
		return item in self.name2idx

	def __getitem__(self, item):
		if self.active is None:
			print('From full')
			runs = self.full_info
		else:
			runs = self.active
		return runs[item]

	def filter_idx(self, *indicies):
		indicies = set(indicies)
		self.set_active([run for i, run in enumerate(self.active) if i in indicies])

	def sort_by(self, criterion='date', reverse=False):
		if criterion == 'date':
			active = sorted(self.active, key=lambda r: r.meta.date, reverse=reverse)
		elif criterion == 'job':
			active = sorted(self.active, key=lambda r: r.meta.job, reverse=reverse)
		elif criterion == 'model':
			active = sorted(self.active, key=lambda r: r.meta.model, reverse=reverse)
		elif criterion == 'data':
			active = sorted(self.active, key=lambda r: r.meta.dataset, reverse=reverse)
		elif criterion == 'ckpt':
			active = sorted(self.active, key=lambda r: r.meta.ckpt, reverse=reverse)
		else:

			active = criterion(self.active)

			# raise Exception('Unknown criterion: {}'.format(criterion))

		self.set_active(active)

		return self

	def filter(self, criterion):
		active = [run for run in self.active if criterion(run)]
		# print('{} remaining'.format(len(self.active)))
		self.set_active(active)
		return self

	def filter_checkpoints(self, num, cname='checkpoint_{}.pth.tar'):

		cname = cname.format(num)

		remaining = []
		for run in self.active:
			if cname in os.listdir(run.path):
				remaining.append(run)

		self.set_active(remaining)
		return self


	# def filter_complete(self, complete=None):
	# 	remaining = self._check_incomplete(complete=complete, show=False, negate=True)
	# 	self.set_active(remaining)
	# 	return self
	# def filter_incomplete(self, complete=None):
	# 	remaining = self._check_incomplete(complete=complete, show=False, negate=False)
	# 	self.set_active(remaining)
	# 	return self

	def filter_sel(self, sel): # for slices
		self.set_active(self.active[sel])

	def filter_min(self, date=None, job=None, ckpt=None):

		if job is not None:
			self.sort_by('job')
			jobs = sorted([run.meta.job[0] for run in self.active])
			idx = bisect_left(jobs, str(job).zfill(4))
			self.set_active(self.active[idx:])

		if date is not None:
			self.sort_by('date')
			dates = sorted([run.meta.date[0] for run in self.active])
			idx = bisect_left(dates, date)
			self.set_active(self.active[idx:])

		if ckpt is not None:
			self.sort_by('ckpt')
			ckpts = sorted([run.meta.ckpt for run in self.active])
			idx = bisect_left(ckpts, ckpt)
			self.set_active(self.active[idx:])

		return self

	def filter_dates(self, *dates):
		remaining = []
		dates = set(dates)

		for run in self.active:
			day, time = run.meta.date
			if day in dates:
				remaining.append(day)

		self.set_active(remaining)
		return self

	def filter_jobs(self, *jobs):
		remaining = []
		allowed = {str(job) for job in jobs}

		filled = set()
		for job in allowed:
			if '-' in job:
				j,p = job.split('-')
				filled.add('{}-{}'.format(j.zfill(4),p.zfill(2)))
		allowed.update(filled)

		allowed.update({job.zfill(4) for job in allowed})

		for run in self.active:
			job, sch, proc = run.meta.job
			long_job = '{}-{}'.format(job, proc)
			# print(long_job)
			if job in allowed or long_job in allowed:
				remaining.append(run)

		self.set_active(remaining)
		return self

	def filter_strs(self, *strs):
		remaining = []
		for run in self.active:
			for s in strs:
				if ('!' == s[0] and s[1:] not in run.name) or s in run.name:
					remaining.append(run)

		self.set_active(remaining)
		return self

	def filter_models(self, *models):
		remaining = []
		models = set(models)
		for run in self.active:
			model = run.meta.model
			if model in models:
				remaining.append(run)

		self.set_active(remaining)
		return self

	def filter_datasets(self, *datasets):
		remaining = []
		datasets = set(datasets)
		for run in self.active:
			data = run.meta.dataset
			if data in datasets:
				remaining.append(run)

		self.set_active(remaining)
		return self

	def clear_filters(self):
		self.set_active()
		return self

	# def select(self, model=None, dataset=None):
	# 	if model is not None:
	# 		self.filter(lambda r: r.config.info.model_type == model)
	# 	if dataset is not None:
	# 		self.filter(lambda r: r.config.info.dataset_type == dataset)
	# 	return self


	# def options(self, models=True, datasets=True):
	#
	# 	out = []
	#
	# 	if models:
	# 		mopts = set()
	# 		for run in self.active:
	# 			mopts.add(run.config.info.model_type)
	# 		out.append(mopts)
	#
	# 	if datasets:
	# 		dopts = set()
	# 		for run in self.active:
	# 			dopts.add(run.config.info.dataset_type)
	# 		out.append(dopts)
	#
	# 	if len(out) == 1:
	# 		return out[0]
	# 	return out

	def _torun(self, x):
		if x is not None and not isinstance(x, Run):
			x = self(x) if isinstance(x, str) else self[x]
		return x

	def compare(self, base, other=None, bi=False, ignore_keys=None):

		base, other = self._torun(base), self._torun(other)

		if 'config' not in base:
			self.load_configs()

		protected_keys = self._extended_protected_keys.copy() if other is None else self._protected_keys.copy()

		if other is None:
			other = base.config
			base = base.base
		else:
			base, other = base.config, other.config

		if ignore_keys is not None:
			protected_keys.update(ignore_keys)

		return compare_config(base, other=other, bi=bi, ignore_keys=protected_keys)

	# def show_unique(self, ignore_keys=None):
	#
	# 	for i, run in enumerate(self.active):
	#
	# 		print('{:>3}) {}'.format(i, run.name))
	#
	# 		if 'diffs' not in run:
	# 			run.diffs = self.compare(run, ignore_keys=ignore_keys)
	# 		diffs = run.diffs
	# 		base = run.base
	#
	# 		for ks in util.flatten_tree(diffs):
	# 			print('{}{} - {} ({})'.format('\t', '.'.join(map(str, ks)), diffs[ks], (base[ks] if ks in base else '_')))#, util.deep_get(diffs, ks)))
	# 		print()
	#
	# 		run.diffs = diffs

	def _get_unique(self, run, ignore_keys=None):
		if 'diffs' not in run:
			run.diffs = self.compare(run, ignore_keys=ignore_keys)
		diffs = run.diffs
		base = run.base

		unique = ['{} - {} ({})'.format('.'.join(map(str, ks)), diffs[ks],
			                              (base[ks] if ks in base else '_'))
		          for ks in util.flatten_tree(diffs)]

		return unique

	def show(self, *props, ignore_keys=None, indent='\t', manuals=[]):

		props = set(props)

		for i, run in enumerate(self.active):
			print('{:>3}) {}'.format(i, run.name))

			if 'all' in props or 'checkpoint' in props or 'ckpt' in props:
				est = ' (approx {} steps)'.format(run.config.output.save_freq * run.meta.ckpt) if 'config' in run else ''
				print('{}Checkpoint: {}{}'.format(indent, run.meta.ckpt, est))

			if ('all' in props or 'steps' in props) and 'records' in run:
				print('{}Steps: {}'.format(indent, run.records['total_steps']))

			if ('all' in props or 'stats' in props) and 'stats' in run:
				stats = ['{:>10} : {:3.2f}'.format(*item) for item in run.stats.items()]
				print('\n'.join(['{}{}'.format(indent, line) for line in stats]))
				#print()

			if 'all' in props or 'unique' in props:
				unique = self._get_unique(run, ignore_keys=ignore_keys)
				print('\n'.join(['{}{}'.format(indent, line) for line in unique]))

			if len(manuals):
				info = []
				for manual in manuals:
					info.extend(manual(run).split('\n'))
				print('\n'.join(['{}{}'.format(indent, line) for line in info]))

			if len(props) or len(manuals):
				print()


	def start_tb(self, port=None):
		assert self.tbout is not None, 'no tbout'

		if self.tb is None:

			argv = [None, '--logdir', self.tbout]
			if port is not None:
				argv.extend(['--port', str(port)])

			tb = program.TensorBoard()
			tb.configure(argv=argv)
			self.tb_url = tb.launch()
			self.tb = tb
		print('Tensorboard started: {}'.format(self.tb_url))

	def clear_links(self):
		assert self.tbout is not None, 'no tbout'
		for name in os.listdir(self.tbout):
			os.unlink(os.path.join(self.tbout, name))

		for run in self.active:
			for fname in os.listdir(run.path):
				full = os.path.join(run.path, fname)
				if os.path.islink(full):
					os.unlink(full)

	def _val_format(self, val):
		if isinstance(val, (tuple, list)):
			return '{}=' + ','.join(map(str,val))
		if isinstance(val, float):
			return '{}=' + '{:.2g}'.format(val).replace('.', 'p')
		if val == '__removed__':
			return 'no-{}'
		return '{}=' + str(val)

	def link(self, name_fmt='{name}'):

		assert self.tbout is not None, 'no tbout'

		# assert include_dataset or include_date or include_diffs \
		#        or include_idx or include_job or include_model, 'no name for links'

		for i, run in enumerate(self.active):

			unique = None
			if 'unique' in name_fmt:
				if 'diffs' not in run:
					run.diffs = self.compare(run)

				if len(run.diffs):

					unique = '__'.join(self._val_format(run.diffs[ks]).format('.'.join(ks[1:]))
					                 for ks in util.flatten_tree(run.diffs))

				else:
					unique = 'default'

			if 'date' in run.config.info:
				date = run.config.info.date
			elif isinstance(run.config.output._logged_date, str):
				date = run.config.output._logged_date
			else:
				date = run.name.split('_')[-1]

			name = name_fmt.format(idx=i,
			                       unique=unique,
			                       name=run.name,
			                       model=run.config.info.model_type,
			                       dataset=run.config.info.dataset_type,
			                       date=date,
			                       job=run.config.name.split('_')[-1])

			link_path = os.path.join(self.tbout, name)

			if os.path.islink(link_path): # avoid duplicate links
				os.unlink(link_path)
			if 'link' in run and os.path.islink(run.link): # avoid duplicate models
				os.unlink(run.link)

			os.system('ln -s {} {}'.format(run.path, link_path))
			run.link = link_path


# class Run(util.adict):
#
# 	def clear(self):
# 		if 'state' in self:
# 			del self.state
# 			torch.cuda.empty_cache()
#
# 	def reset(self, state=None):
# 		self.clear()
#
# 		if state is None:
# 			state = util.adict()
#
# 		if 'ckpt_path' not in state:
# 			state.ckpt_path = self.ckpt_path
#
# 		self.state = state
# 		return state
#
# 	def load(self, **kwargs):
# 		if 'state' not in self:
# 			self.reset()
#
# 		self._manager._load_fn(self.state, **kwargs)
#
# 	def load_into(self, name, key=None):
# 		if key is None:
# 			key = name.split('.')[0]
# 		if key not in self:
# 			path = os.path.join(self.path, name)
# 			if not os.path.isfile(path):
# 				print('{} not found in: {}'.format(name, self.name))
# 			else:
# 				self[key] = torch.load(path)
#
# 	def load_config(self, force=False, clear_info=False, excluded_info=[]):
# 		if 'config' not in self or force:
#
# 			fname = os.path.join(self.path, 'config.yml')
#
# 			if not os.path.isfile(fname):
# 				print('{} has no config'.format(self.name))
# 				return
#
# 			config = get_config(fname)
#
# 			if clear_info:
# 				del self.config.info
# 			elif 'info' in config:
#
# 				for k, v in config.info.items():
# 					if k in excluded_info:
# 						pass
# 					elif 'dataset_type' == k:
# 						new = v
# 						# if 'dataset' in self.meta:
# 						# 	assert self.meta.dataset == new, '{} vs {}'.format(self.meta.dataset, new)
# 						self.meta.dataset = new
# 					elif 'model_type' == k:
# 						new = v
# 						# if 'model' in self.meta:
# 						# 	assert self.meta.model == new, '{} vs {}'.format(self.meta.model, new)
# 						self.meta.model = new
# 					elif 'history' == k:
# 						self.meta.history = config.info.history
# 					elif 'date' == k:
# 						new = tuple(v.split('-'))
# 						# if 'date' in self.meta:
# 						# 	assert self.meta.date == new, '{} vs {}'.format(self.meta.date, new)
# 						self.meta.date = new
# 					else:
# 						self.meta[k] = v
# 			if 'job' in config:
# 				self.meta.job = config.job.num, config.job.ID, config.job.ps
#
# 			try:
# 				base = get_base_config(config)
# 			except Exception:
# 				# print('WARNING: loading base failed')
# 				base = None
#
# 			if base is None:
# 				print('{} has no base'.format(self.name))
# 				return
#
# 			self.config = config
# 			self.base = base
#
# 	def __str__(self):
# 		return self.name
#
# 	def __repr__(self):
# 		return 'Run({})'.format(self.name)
#
# 	def run(self, **kwargs):
# 		self._manager._run_model_fn(self.state, **kwargs)
#
# 	def evaluate(self, force=False, pbar=None):
#
# 		jobs = self._manager._eval_fns.items()
# 		# if pbar is not None:
# 		# 	jobs = pbar(jobs, total=len(self._manager._eval_fns))
#
# 		if 'evals' not in self.state:
# 			self.state.evals = {}
# 		results = self.state.evals
#
# 		for k, fn in jobs:
# 			if k not in results or force:
# 				print('--- Evaluating: {}'.format(k))
# 				start = time.time()
# 				# if pbar is not None:
# 				# 	jobs.set_description('EVAL: {}'.format(k))
# 				results[k] = fn(self.state, pbar=pbar)
# 				print('... took: {:3.2f}'.format(time.time() - start))
# 			print('{}: {}\n'.format(k, results[k]))
#
# 		# self.state.evals = results
# 		return results
#
# 	def visualize(self, pbar=None):
#
# 		jobs = self._manager._viz_fns.items()
# 		if pbar is not None:
# 			jobs = pbar(jobs, total=len(self._manager._viz_fns))
#
# 		results = {}
# 		for k, fn in jobs:
# 			start = time.time()
# 			if pbar is None:
# 				print('--- Visualizing: {}'.format(k))
# 			else:
# 				jobs.set_description('VIZ: {}'.format(k))
#
# 			results[k] = [Visualization(fig) for fig in fn(self.state, pbar=pbar)]
#
# 			if pbar is None:
# 				print('... took: {:3.2f}\n'.format(time.time() - start))
#
# 		if pbar is not None:
# 			jobs.close()
#
# 		self.state.figs = results
# 		return results
#
# 	def save(self, save_dir=None, fmtdir='{}', overwrite=False, append_ckpt=True,
# 	         include_checkpoint=True, include_config=True, include_original=True, include_results=True,
# 	         img_ext='png', vid_ext='mp4'):
#
# 		if save_dir is None:
# 			save_dir = self._manager.save_dir
#
# 		name = self.name
# 		if append_ckpt and 'ckpt' in self.meta and 'ckpt' not in name:
# 			num = self.meta.ckpt
# 			name = '{}_ckpt{}'.format(name, num)
#
# 		save_path = os.path.join(save_dir, fmtdir.format(name))
#
# 		print('Saving results to: {}'.format(save_path))
#
# 		try:
# 			if overwrite:
# 				util.create_dir(save_path)
# 			else:
# 				os.makedirs(save_path)
# 		except FileExistsError:
# 			print('ERROR: File already exists, you can overwrite using the "overwrite" arg')
# 		else:
# 			if include_checkpoint:
# 				src = self.ckpt_path
# 				dest = os.path.join(save_path, 'model.pth.tar')
# 				shutil.copyfile(src, dest)
# 				print('\tModel saved')
# 			if include_config:
# 				src = os.path.join(self.path, 'config.yml')
# 				dest = os.path.join(save_path, 'config.yml')
# 				shutil.copyfile(src, dest)
# 				print('\tConfig saved')
# 			if include_original:
# 				with open(os.path.join(save_path, 'original_ckpt_path.txt'), 'w') as f:
# 					f.write(self.ckpt_path)
#
# 			if 'state' in self:
# 				if 'figs' in self.state:
# 					for name, figs in self.state.figs.items():
# 						if len(figs) == 1:
# 							path = os.path.join(save_path, name)
# 							figs[0].save(path, img_ext=img_ext, vid_ext=vid_ext)
# 						else:
# 							for i, fig in enumerate(figs):
# 								path = os.path.join(save_path, '{}{}'.format(name, str(i).zfill(3)))
# 								fig.save(path, img_ext=img_ext, vid_ext=vid_ext)
#
# 					print('\tVisualization saved')
#
# 				if 'evals' in self.state:
# 					with open(os.path.join(save_path, 'eval.txt'), 'w') as f:
# 						for k, v in self.state.evals.items():
# 							f.write('{} : {}\n'.format(k, str(v)))
# 					torch.save(self.state.evals, os.path.join(save_path, 'eval.pth.tar'))
# 					print('\tEvaluation saved')
#
# 				if include_results and 'results' in self.state:
# 					torch.save(self.state.results, os.path.join(save_path, 'results.pth.tar'))
# 					print('\tResults saved: {}'.format(', '.join(map(str, self.state.results.keys()))))
#
# 		return save_path

#
# def compare_config(base, other=None, bi=False, ignore_keys=None):
# 	# vs_default = False
#
# 	protected_keys = {'name', 'info', 'job', 'save_dir'}
# 	if other is None:
#
# 		protected_keys.update({
# 		'_logged_date', 'din', 'dout', 'info', 'dataroot', 'saveroot', 'run_mode', 'dins', 'douts',
# 		})
#
# 		if 'history' not in base.info:
# 			raise Exception('Unable to load default - no history found')
#
# 		other = base
# 		base = get_base_config(base)
#
# 	if ignore_keys is not None:
# 		protected_keys.update(ignore_keys)
#
# 	diffs = ConfigDict()#util.NS()
# 	_compare_configs(base, other, diffs=diffs, protected_keys=protected_keys)
#
# 	if bi:
# 		adiffs = ConfigDict()#util.NS()
# 		_compare_configs(other, base, diffs=adiffs, protected_keys=protected_keys)
# 		return diffs, adiffs
#
# 	return diffs
#
# def get_base_config(base):
# 	if 'info' in base and 'history' in base.info:
# 		return parse_config(base.info.history)
#
# def _compare_configs(base, other, diffs, protected_keys=None):
#
# 	for k,v in base.items():
#
# 		if k in protected_keys:
# 			pass
# 		elif k not in other:
# 			diffs[k] = '__removed__'
# 		elif isinstance(v, ConfigDict):
# 			_compare_configs(v, other[k], diffs=diffs[k], protected_keys=protected_keys)
# 			if len(diffs[k]) == 0:
# 				del diffs[k]
# 		elif v != other[k]:
# 			diffs[k] = other[k]
#
# 	for k,v in other.items():
# 		if k not in protected_keys and k not in base:
# 			diffs[k] = v
#
#
#
# def render_format(raw, unfolded=False):
# 	unfolded = True
# 	if isinstance(raw, set):
# 		itr = dict()
# 		for i, el in enumerate(raw):
# 			itr['s{}'.format(i)] = render_format(el, unfolded)
# 		return itr
# 	elif isinstance(raw, dict):
# 		return dict((str(k), render_format(v, unfolded)) for k, v in raw.items())
# 	elif isinstance(raw, list):
# 		return list(render_format(el) for el in raw)
# 		itr = dict()
# 		for i, el in enumerate(raw):
# 			itr['l{}'.format(i)] = render_format(el, unfolded)
# 		return itr
# 	elif isinstance(raw, tuple):
# 		return list(render_format(el) for el in raw)
# 		itr = dict()
# 		for i, el in enumerate(raw):
# 			itr['t{}'.format(i)] = render_format(el, unfolded)
# 		return itr
# 	return str(raw)
#
#
# import uuid
# from IPython.display import display_javascript, display_html
#
#
# class render_dict(object):
# 	def __init__(self, json_data):
# 		self.json_str = render_format(json_data)
#
# 		# if isinstance(json_data, dict):
# 		#     self.json_str = json_data
# 		#     #self.json_str = json.dumps(json_data)
# 		# else:
# 		#     self.json_str = json
# 		self.uuid = str(uuid.uuid4())
#
# 	def _ipython_display_(self):
# 		display_html('<div id="{}" style="height: 600px; width:100%;"></div>'.format(self.uuid),
# 		             raw=True
# 		             )
# 		display_javascript("""
# 		require(["https://rawgit.com/caldwell/renderjson/master/renderjson.js"], function() {
# 		  renderjson.set_show_to_level(1)
# 		  document.getElementById('%s').appendChild(renderjson(%s))
# 		});
# 		""" % (self.uuid, self.json_str), raw=True)










from collections import deque
import torch

from torch.utils.data import DataLoader
from torch.utils.data.dataloader import _SingleProcessDataLoaderIter, _MultiProcessingDataLoaderIter
# from .collectors import Batchable_Dataset, Device_Dataset # TODO: clean up import order

from .collectors import Batchable

from .. import util

class Featured_DataLoaderIter:
	
	def __init__(self, loader):
		# loader.batched = isinstance(loader.dataset, Batchable)
		super().__init__(loader)
		# self._dataset_fetcher.auto_collation = not isinstance(loader.dataset, Batchable)
		self.device = loader.device
		self.N = len(loader)
	
	def skip(self, num):
		for _ in range(num):
			self._next_index()
	
	def __next__(self):
		return util.to(super().__next__(), self.device)

class Featured_SingleProcessIter(Featured_DataLoaderIter, _SingleProcessDataLoaderIter):
	pass
class Featured_MultiProcessIter(Featured_DataLoaderIter, _MultiProcessingDataLoaderIter):
	pass

class Featured_DataLoader(DataLoader):
	
	def __init__(self, dataset, *args, device=None, seed=None, generator=None, **kwargs):

		if seed is not None and generator is None:
			generator = torch.Generator()
			generator.manual_seed(seed)

		self.batched = not isinstance(dataset, Batchable)
		# self.batched = None

		super().__init__(dataset, *args, generator=generator, **kwargs)
		
		# self._auto_collation = False

		if device is None:
			try:
				device = self.dataset.device
			except AttributeError:
				device = 'cpu'
		self.device = device
	
	@property
	def _index_sampler(self):
		# The actual sampler used for generating indices for `_DatasetFetcher`
		# (see _utils/fetch.py) to read data at each time. This would be
		# `.batch_sampler` if in auto-collation mode, and `.sampler` otherwise.
		# We can't change `.sampler` and `.batch_sampler` attributes for BC
		# reasons.
		if self.batch_sampler is not None:
			return self.batch_sampler
		else:
			return self.sampler
		
	@property
	def _auto_collation(self):
		return (self.batched is not None and self.batched) or (self.batched is None and super()._auto_collation)
		
	# def _auto_collation(self):
	# 	return not self.batched or super()._auto_collation()
		
	
	# def __iter__(self):
	# 	itr = super().__iter__()
	#
	# 	def _skip(num):
	# 	itr.skip = _skip
	#
	# 	def _move_to():
	# 		batch = next(itr)
	# 		return util.to(batch, self.device)
	# 	itr.__next__ = _move_to
	#
	# 	return itr

	def get_dataset(self):
		return self.dataset

	def get_batch_size(self):
		return self.batch_size

	def __iter__(self):
		if self.num_workers == 0:
			return Featured_SingleProcessIter(self)
		else:
			return Featured_MultiProcessIter(self)



class BatchedDataLoader(Featured_DataLoader): # loads full batches at a time (dataset must be Batched

	def __init__(self, dataset, batch_size, shuffle=True, drop_last=False,
	             auto_reset=False, num_workers=None, pin_memory=False, device=None, **kwargs):
		self.dataset = dataset

		if len(self.dataset) < batch_size:
			print(f'WARNING: batchsize changed from {batch_size} to {len(self.dataset)}')
			batch_size = len(self.dataset)

		# assert len(self.dataset) > batch_size, 'dataset is not large enough for a single batch: {} vs {}'.format(len(dataset), batch_size)

		# if not isinstance(dataset, Batchable_Dataset):
		# 	assert dataset.allow_batched(), 'this dataset doesnt seem to be compatible with a BatchedDataLoader: {}'.format(dataset)

		self.batch_size = batch_size
		self.shuffle = shuffle
		self.num_workers = num_workers
		self.pin_memory = pin_memory
		self.drop_last = drop_last
		self.auto_reset = False
		self.num = len(self.dataset) // batch_size
		if self.num*batch_size != len(self.dataset) and not self.drop_last:
			self.num += 1

		if device is None:
			try:
				device = dataset.device
			except AttributeError:
				device = 'cpu'
		self.device = device
		# self.device = dataset.device if isinstance(dataset, Device_Dataset) else 'cpu'

	def __len__(self):
		return self.num

	def __iter__(self):
		return _BatchedDataLoaderIter(self.dataset, self.batch_size, self.shuffle, self.drop_last, self.auto_reset, self.device)

class _BatchedDataLoaderIter(object):
	def __init__(self, dataset, batch_size, shuffle=True, drop_last=False, auto_reset=False, device=None):
		self.dataset = dataset
		self.batch_size = batch_size
		self.shuffle = shuffle
		self.drop_last = drop_last
		self.auto_reset = auto_reset
		self.device = device

		self.remainder = None
		self.batches = []
		self.idx = None
		self.reset()

	def reset(self):

		if len(self.dataset) < 1e7:

			order = torch.randperm(len(self.dataset), device=self.device) if self.shuffle else torch.arange(0, len(self.dataset), device=self.device)

			history = None
			if self.remainder is not None:
				pull = self.batch_size - len(self.remainder)
				history = torch.cat([self.remainder, order[:pull]])
				order = order[pull:]

				self.remainder = None

			batches = list(torch.split(order, self.batch_size))

			if history is not None:
				batches.insert(0,history)

			last = batches[-1]
			if len(last) < self.batch_size:
				if self.drop_last or self.auto_reset:
					batches.pop()
					if self.auto_reset:
						self.remainder = last

			self.batches = batches

		else:
			self.batches = None
		self.idx = 0


	def skip(self, num):
		self.idx += num

	def __len__(self):
		return len(self.dataset)//self.batch_size if self.batches is None else len(self.batches) - self.idx

	# def next_batch(self):
	# 	batch = next(self)
	# 	return util.to(batch, self.device)

	def __iter__(self):
		return self

	def __next__(self):

		if self.batches is None:

			idx = torch.randint(len(self.dataset), size=(self.batch_size,))
			batch = self.dataset[idx]

		else:

			if self.idx >= len(self.batches):
				if self.auto_reset:
					self.reset()
				else:
					raise StopIteration

			batch = self.dataset[self.batches[self.idx]]
			self.idx += 1

		return util.to(batch, self.device)
		# return batch

	def __getitem__(self, item):
		return self.dataset[self.batches[item]]


def get_loaders(*datasets, batch_size=64, num_workers=0, shuffle=True, pin_memory=True,
		   drop_last=False, worker_init_fn=None, allow_batched=True, device='cpu'):

	if shuffle == 'all':
		shuffles = [True]*3
	elif shuffle:
		shuffles = [True, False, False]
	else:
		shuffles = [False]*3

	for ds in datasets:
		if ds is not None:
			break
	if ds is None:
		return datasets if len(datasets) > 1 else None # all are None

	loader_cls = Featured_DataLoader
	kwargs = {
		'batch_size': batch_size,
		'drop_last': drop_last,
		'device': device,
	}

	if allow_batched:
		try:
			assert ds.allow_batched()
		except (AttributeError, AssertionError):
			pass
		else:
			# print('Using batched data loader')
			loader_cls = BatchedDataLoader
	else:

		try:
			assert ds.get_device() == 'cpu'
		except AttributeError:
			pass
		except AssertionError:
			pin_memory = False

		kwargs.update({
			'pin_memory': pin_memory,
			'worker_init_fn': worker_init_fn,
			'num_workers': num_workers,
		})


	loaders = [(loader_cls(ds, shuffle=s, **kwargs) if ds is not None else None)
	           for ds, s in zip(datasets, shuffles)]

	# if not silent: # TODO: deprecated!
	# 	trainloader = loaders[0]
	# 	testloader = None if len(loaders) < 2 else loaders[-1]
	# 	valloader = None if len(loaders) < 3 else loaders[1]
	#
	# 	print('traindata len={}, trainloader len={}'.format(len(datasets[0]), len(trainloader)))
	# 	if valloader is not None:
	# 		print('valdata len={}, valloader len={}'.format(len(datasets[1]), len(valloader)))
	# 	if testloader is not None:
	# 		print('testdata len={}, testloader len={}'.format(len(datasets[-1]), len(testloader)))
	# 	print('Batch size: {} samples'.format(batch_size))

	if len(loaders) == 1:
		return loaders[0]
	return loaders





from .collectors import *
from .register import Dataset, dataset_registry
from . import manager
from .loaders import get_loaders, DataLoader, BatchedDataLoader
from . import preprocess
from .samplers import Intervention_Sampler, JointFactorSampler
# from .samplers import Generator




import omnifig as fig

import torch
from torch import nn

from .. import util
from . import framework as fm


def load_checkpoint(path):  # model parameters - TODO: maybe add more options
	try:
		return torch.load(path)
	except RuntimeError:
		return torch.load(path, map_location = "cpu")


@fig.Script('load-model', description='Creates/loads a model')
@fig.Component('model')
def load_model(A):
	'''
	Creates the model and possibly loads existing model parameters
	'''
	
	raw_type = A.pull('_type', silent=True)
	if raw_type == 'model':
		model_type = A.pull('_model_type', None, silent=True)
		if model_type is None:
			assert A.contains_nodefault('model')
			A = A.sub('model')
		else:
			A.push('_type', model_type, silent=True)
			A.push('_mod', A.pull('_model_mod', []), silent=True)
	
	util.Seed(A)
	
	# model = fig.create_component(A)
	model = A.pull_self()
	
	optim = model.set_optim(A) if isinstance(model, fm.Optimizable) else None
	
	model.to(model.get_device())
	
	print_model = A.pull('_print_model', False)
	if print_model:
		print(model)
		if A.pull('_print_optim', False):
			print(optim)
		print(f'Number of model parameters: {util.count_parameters(model)}')
	
	path = A.pull('_load-ckpt', None)
	if path is not None:
		model.load_checkpoint(path)
		if print_model:
			print(f'Loaded parameters from {path}')
	
	# origin_name = A.pull('__origin_key', None, silent=True)
	
	return model



fig.AutoComponent('criterion')(util.get_loss_type)
fig.AutoComponent('nonlin')(util.get_nonlinearity)
fig.AutoComponent('normalization')(util.get_normalization)
fig.AutoComponent('pooling')(util.get_pooling)
fig.AutoComponent('upsampling')(util.get_upsample)

@fig.AutoComponent('viz-criterion')
class Viz_Criterion(nn.Module):
	def __init__(self, criterion, arg_names=[], kwarg_names=[],
	             allow_grads=False):
		super().__init__()
		
		self.criterion = util.get_loss_type(criterion)
		self.arg_names = arg_names
		self.kwarg_names = kwarg_names
		self.allow_grads = allow_grads

	def forward(self, out):
		
		args = [out[key] for key in self.arg_names]
		kwargs = {key:out[key] for key in self.kwarg_names}
		
		if self.allow_grads:
			return self.criterion(*args, **kwargs)
		
		with torch.no_grad():
			return self.criterion(*args, **kwargs)

# @Component('stage') # TODO
class Stage_Model(fm.Model):
	def __init__(self, A):
		stages = A.pull('stages')

		din = A.pull('din')
		dout = A.pull('dout')

		criterion_info = A.pull('criterion', None)

		super().__init__(din, dout)

		self.stages = nn.ModuleList(self._process_stages(stages))

		self.criterion = None
		if criterion_info is not None:
			self.criterion = util.get_loss_type(criterion_info) \
				if isinstance(criterion_info, str) else util.get_loss_type(**criterion_info)  # TODO: automate



	def _create_stage(self, stage):
		return fig.create_component(stage)
	def _process_stages(self, stages):

		N = len(stages)

		assert N > 0, 'no stages provided'

		din = self.din

		sub = []
		for i, stage in enumerate(stages):

			stage.din = din
			stage = self._create_stage(stage)

			sub.append(stage)
			din = stage.dout

		assert din == self.dout, 'dins and douts not set correctly: {} vs {}'.format(din, self.dout)

		return sub

	def forward(self, x):
		q = x
		for stage in self.stages:
			q = stage(q)
		return q


	def _step(self, batch, out=None):
		if out is None:
			out = util.TensorDict()

		# compute loss

		x = batch[0]
		if len(batch) == 2:
			y = batch[1]
		else:
			assert len(batch) == 1, 'Cant figure out how the batch is setup (you will have to implement your own _step)'
			y = x

		out.x, out.y = x, y

		pred = self(x)
		out.pred = pred

		loss = self.criterion(pred, y)
		out.loss = loss

		if isinstance(self, fm.Regularizable):
			reg = self.regularize(out)
			loss += reg

		out.loss = loss

		if self.train_me():
			self.optim.zero_grad()
			loss.backward()
			self.optim.step()

		return out








import time
from pathlib import Path
from collections import OrderedDict
from omnibelt import load_json, save_json
import omnifig as fig

from .. import util
from ..util import Named, StatsClient, Configurable, Statelike, Checkpointable

################
# region Clocks
################

class AlertNotFoundError(Exception):
	def __init__(self, name):
		super().__init__(f'Could not find an alert named {name}')
		self.name = name

@fig.Component('clock/simple')
class SimpleClock(Checkpointable, Configurable):
	def __init__(self, A, skip_load=None, **kwargs):
		
		if skip_load is None:
			skip_load = A.pull('skip_load', False)
		
		ckpt = None
		if not skip_load:
			ckpt = A.pull('_load-ckpt', None)
		
		super().__init__(A, **kwargs)
		self.ticks = 0
		self.alerts = OrderedDict()
		self._info = None
		
		self._skip_load = None
		
		if ckpt is not None:
			self.load_checkpoint(ckpt)
	
	def prep(self, info=None):
		if info is None:
			info = self._info
		for idx, (name, alert) in enumerate(self.alerts.items()):
			alert.prep(idx, info=info)
	
	def sort_alerts(self, start_with=None, end_with=None, strict=True):
		order = OrderedDict()
		
		if start_with is not None:
			for name in start_with:
				if name in self.alerts:
					order[name] = self.alerts[name]
				elif strict:
					raise AlertNotFoundError(name)
				
		for name, alert in self.alerts.items():
			if end_with is None or name not in end_with:
				order[name] = alert
		
		if end_with is not None:
			for name in end_with:
				if name in self.alerts:
					order[name] = self.alerts[name]
				elif strict:
					raise AlertNotFoundError(name)
		
		self.alerts.clear()
		self.alerts.update(order)
	
	def get_info(self):
		return self._info
	
	def set_info(self, info):
		self._info = info
	
	def register_alert_fn(self, name, check=None, activate=None):
		self.register_alert(name, CustomAlert(check=check, activate=activate))
	
	def register_alert(self, name, alert, **unused):
		if name is None:
			name = f'{alert}#{id(alert)}'
		self.alerts[name] = alert
	
	def checkpoint(self, path, ident='clock'):
		
		path = Path(path) / f'{ident}.json'
		
		states = {}
		for name, alert in self.alerts.items():
			if isinstance(alert, Savable):
				states[name] = alert.state_dict()
			
		save_json({'time': self.get_time(), 'alert_states': states}, path)
	
	def load_checkpoint(self, path, ident='clock'):
		
		path = Path(path) / f'{ident}.json'
		
		if path.is_file():
			data = load_json(path)
			
			self.ticks = data['time']
			
			states = data['alert_states']
			for name, state in states.items():
				if name in self.alerts:
					self.alerts[name].load_state_dict(state)
	
	def clear(self):
		self.alerts.clear()
		self._info = None
	
	def _call_alert(self, name, alert, info=None):
		if info is None:
			info = self._info
		return alert.activate(self.ticks, info)
	
	def get_time(self):
		return self.ticks
	
	def __len__(self):
		return self.get_time()
	
	def set_time(self, ticks):
		self.ticks = ticks
	
	def tick(self, info=None):
		self.ticks += 1
		for name, alert in self.alerts.items():
			if alert.check(self.ticks, info):
				self._call_alert(name, alert, info=info)
	
	def step(self, info=None, n=None):
		if n is None:
			n = 1
		for _ in range(n):
			self.tick(info=info)


@fig.AutoModifier('clock/limit')
class Limited(SimpleClock):
	def __init__(self, A, **kwargs):
		limit = A.pull('limit', None)
		super().__init__(A, **kwargs)
		self.limit = limit

	def get_limit(self):
		return self.limit
	
	def get_remaining(self):
		lim = self.get_limit()
		if lim is not None:
			return max(lim-self.get_time(), 0)
	
	def set_limit(self, limit):
		old = self.limit
		self.limit = limit
		return old
	
	def tick(self, info=None, force=False):
		if self.get_limit() < self.get_time() and not force:
			raise StopIteration
		return super().tick(info=info)
	
	def step(self, info=None, n=None):
		if n is None:
			n = self.get_remaining()
		return super().step(info=info, n=n)


@fig.AutoModifier('clock/stats')
class Stats(StatsClient, SimpleClock):
	# def __init__(self, A, **kwargs):
	# 	super().__init__(A, **kwargs)
	
	
	def register_alert(self, name, alert, add_to_stats=True, **unused):
		if add_to_stats:
			self.register_stats(name)
		return super().register_alert(name, alert, **unused)
	
	def _call_alert(self, name, alert, info=None):
		val = super()._call_alert(name, alert, info=info)
		if val is not None:
			self.mete(name, val)
		return val


@fig.AutoModifier('clock/timed')
class Timed(Stats):
	def __init__(self, A):
		super().__init__(A)
		self._timed_fmt = A.pull('timed-fmt', 'time-{}')
		self._timed_stats = {}
	
	def register_alert(self, name, alert, add_to_stats=True, add_to_timed=True):
		if add_to_timed:
			key = self._timed_fmt.format(name)
			self._timed_stats[name] = key
			self.register_stats(key)
		super().register_alert(name, alert, add_to_stats=add_to_stats)
	
	def _call_alert(self, name, alert, info=None):
		start = time.time()
		
		out = super()._call_alert(name, alert, info=info)
		
		if name in self._timed_stats:
			self.mete(self._timed_stats[name], time.time() - start)
		
		return out

@fig.Component('clock')
class Clock(Timed, Limited):
	pass


# endregion
################


################
# region Alerts
################


class AlertBase:
	def prep(self, order, info=None):
		pass
	
	def check(self, tick, info=None):
		return True
	
	def activate(self, tick, info=None):
		'''

		:param tick: int
		:param info: object passed to clock.tick()
		:return: new value (representative of the alert) or none
		'''
		pass


class CustomAlert(AlertBase):
	def __init__(self, activate=None, check=None, **kwargs):
		super().__init__(**kwargs)
		self._activate = activate
		self._check = check
	
	def check(self, tick, info=None):
		if self._check is None:
			return True
		return self._check(tick, info=info)
	
	def activate(self, tick, info=None):
		if self._activate is None:
			pass


class Alert(Configurable, AlertBase):
	pass



@fig.AutoModifier('alert/reg')
class Reg(Named, Alert):
	def __init__(self, A, **kwargs):
		super().__init__(A, **kwargs)
		clock = A.pull('clock', None, ref=True)
		if clock is not None:
			clock.register_alert(self.get_name(), self)


@fig.AutoModifier('alert/freq')
class Freq(Alert):
	def __init__(self, A, **kwargs):
		zero = A.pull('include_zero', False)
		freq = A.pull('freq', None)
		super().__init__(A, **kwargs)
		self.freq = freq
		self._include_zero = zero
	
	def check(self, tick, info=None):
		return self.freq is None or ((self._include_zero or tick >= 1) and tick % self.freq == 0)

class Savable(Statelike, Alert):
	pass

# endregion
################


import torch
from torch import nn

import omnifig as fig

# from .. import framework as fm
from ..op import framework as fm
from .. import util


@fig.AutoComponent('multigroup-cls')
class MultiGroupClassification(fm.FunctionBase):
	def __init__(self, group_sizes, group_weights=None):
		super().__init__(sum(group_sizes), 1)
		
		self.sizes = group_sizes
		
		self.criterion = nn.CrossEntropyLoss()
		
		if group_weights is not None:
			raise NotImplementedError
		
	def extra_repr(self) -> str:
		return 'groups = {}'.format(str(self.sizes))
		
	def get_sizes(self):
		return self.sizes
	
	def __len__(self):
		return len(self.sizes)
		
	def forward(self, pred, lbls):
		loss = 0
		for i, w in enumerate(self.split(pred)):
			y = lbls[:,i]
			loss += self.criterion(w, y)
		return loss
		
	def split(self, pred):
		return pred.split(self.sizes, dim=1)
	
	def _get_info(self, pred):
		return zip(*[s.max(dim=1) for s in self.split(pred)])

	def get_info(self, pred):
		return [torch.stack(info, dim=1) for info in self._get_info(pred)]

	def get_confidences(self, pred):
		return torch.stack(self._get_info(pred)[0], dim=1)
	
	def get_picks(self, pred):
		return torch.stack(self._get_info(pred)[1], dim=1)


class Feature_Match(nn.Module):

	def __init__(self, layers, criterion='mse', weights=None,
	             out_wt=None, out_criterion=None,
	             model=None, reduction='mean'):
		super().__init__()

		self._features = []

		def hook(m, input, out):
			self._features.append(out)

		L = 0
		for layer in layers:
			L += 1
			layer.register_forward_hook(hook)
		self.num = L

		if weights is not None:
			assert L == len(weights), '{} != {}'.format(len(layers), len(weights))
			self.weights = weights
			# self.register_buffer('weights', weights)
		else:
			self.weights = None
		self.criterion = util.get_loss_type(criterion)
		assert reduction in {'mean', 'sum', 'none'}
		self.reduction = reduction

		self.out_wt = out_wt
		self.out_criterion = util.get_loss_type(out_criterion)
		if self.out_wt is not None and self.out_wt > 0:
			self.num += 1

		self._model = [model] # dont let pytorch treat this as a submodule

	def __len__(self):
		return self.num

	def clear(self):
		self._features.clear()

	def forward(self, p, q, model=None):

		if model is None:
			model = self._model[0]

		self.clear()
		po = model(p)
		pfs = self._features.copy()

		self.clear()
		qo = model(q)
		qfs = self._features.copy()

		self.clear()

		if self.weights is None:
			losses = [self.criterion(pf, qf) for pf, qf in zip(pfs, qfs)]
		else:
			losses = [w*self.criterion(pf, qf) for w, pf, qf in zip(self.weights, pfs, qfs)]

		if self.out_wt is not None and self.out_wt > 0:
			criterion = self.criterion if self.out_criterion is None else self.out_criterion
			out_loss = self.out_wt * criterion(po, qo)

			losses.append(out_loss)

		if self.reduction == 'none':
			return losses

		loss = sum(losses)

		if self.reduction == 'mean':
			return loss / len(losses)

		return loss







import sys
from pathlib import Path
from tqdm import tqdm

from omnibelt import unspecified_argument

import omnifig as fig

from .. import util

# from .loading import load_config, load_records, setup_logging, setup_records, \
# 	wrap_datasets, wrap_transaction, save_checkpoint
from .loading import respect_config
from .evaluation import evaluate
from .clock import Freq, Reg
from .framework import Visualizable



class RunEvent(Freq):
	def check(self, tick, info=None):
		return self.freq is not None and super().check(tick, info=info)
	
	def purge(self):
		pass

@fig.Component('run/epoch')
class Epoch(RunEvent):
	def __init__(self, A, **kwargs):
		step_limit = A.pull('step-limit', None)
		inline = A.pull('inline', False)
		pbar = None
		if inline:
			pbar = A.pull('pbar', None)
		mode = A.pull('mode', 'val')
		print_results = A.pull('print_metric', True)
		fixed_loader_seed = A.pull('loader-seed', None)

		dataset = A.pull('dataset', None, ref=True)
		model = A.pull('model', None, ref=True)
		records = A.pull('records', None, ref=True)
		
		super().__init__(A, **kwargs)
		
		self.mode = mode
		self.loader_seed = fixed_loader_seed
	
		self.dataset = dataset
		self.model = model
		self.records = records
	
		self.step_limit = step_limit
		# if self.step_limit is not None:
		# 	raise NotImplementedError
		
		self.print_results = print_results
		self.pbar = pbar
		
	def purge(self):
		self.dataset = None
		self.model = None
		self.records = None
		super().purge()
		
	def get_mode(self):
		return self.get_name()
	
	def set_mode(self, mode):
		self.mode = mode
	
	@staticmethod
	def pre_epoch(mode, dataset, model, records=None):
		
		dataset.switch_to(mode)
		model.switch_to(mode)
		if records is not None:
			records.switch_to(mode)
			
			if mode not in records['total_epochs']:
				records['total_epochs'][mode] = 0
			if mode not in records['total_steps']:
				records['total_steps'][mode] = 0
			if mode not in records['total_samples']:
				records['total_samples'][mode] = 0
			
	@staticmethod
	def post_epoch(mode, dataset, model, records=None):
		
		if records is not None:
			if mode not in records['total_epochs']:
				records['total_epochs'][mode] += 1
		
			records['total_epochs'][mode] += 1

		# if self.print_results:
		# 	metric = model.get_metric()
		# 	print('')
	
	@staticmethod
	def epoch_step(batch, model, records=None):
		
		out = model.step(batch)
		if records is not None:
			mode = records.get_mode()
			records['total_samples'][mode] += batch.size(0)
			records['total_steps'][mode] += 1
			
		return out

	@staticmethod
	def epoch_end(dataset, model, records=None, batch=None, out=None):
		
		if out is not None and records is not None:
			records.step()
			if isinstance(model, Visualizable):
				model.visualize(out, records)
		
		name = model.optim_metric
		meter = model.get_metric()
		
		mode = model.get_mode()
		title = mode
		if records is not None:
			ep = records['total_epochs'].get(mode, 0) + 1
			title = f'{mode} ({ep})'
		print(f'{title} {name}: {meter.avg:.4f} (min={meter.min:.4f}, max={meter.max:.4f})')
		print()
		

	def run_epoch(self, dataset=unspecified_argument, model=unspecified_argument, records=unspecified_argument,
	              pre_epoch=unspecified_argument, step_fn=unspecified_argument,
	              epoch_end=unspecified_argument, post_epoch=unspecified_argument):
		
		# region fill in missing args
		
		if pre_epoch is unspecified_argument:
			pre_epoch = self.pre_epoch
		if step_fn is unspecified_argument:
			step_fn = self.epoch_step
		if epoch_end is unspecified_argument:
			epoch_end = self.epoch_end
		if post_epoch is unspecified_argument:
			post_epoch = self.post_epoch
		
		if dataset is unspecified_argument:
			dataset = self.dataset
		if model is unspecified_argument:
			model = self.model
		if records is unspecified_argument:
			records = self.records
		
		# endregion
		
		pre_epoch(self.mode, dataset, model, records)
		
		loader = dataset.get_loader(infinite=False, seed=self.loader_seed)
		N = len(loader)
		if self.step_limit is not None and N > self.step_limit:
			N = self.step_limit
		loader = enumerate(loader)
		if self.pbar is not None:
			loader = self.pbar(loader, limit=N)
		
		if records is not None:
			ep = records['total_epochs'].get(self.mode,0) + 1
			title = f'{self.mode}:{ep}'
		else:
			title = self.mode
		
		batch, out = None, None
		for i, batch in loader:
			out = step_fn(batch, model, records)
			
			if self.pbar is not None:
				progress = model.get_description()
				self.pbar.set_description(f'{title} {progress}')
			
			if self.step_limit is not None and i >= self.step_limit:
				break
		
		if self.pbar is not None:
			self.pbar.reset()
		
		epoch_end(dataset, model, records, batch, out)
		
		post_epoch(self.mode, dataset, model, records)
		
		return batch, out
		
	
	def activate(self, tick=None, info=None):
		
		dataset, model, records = None, None, None
		
		if info is not None:
			dataset = info.get_data() if self.dataset is None else self.dataset
			model = info.get_model() if self.model is None else self.model
			records = info.get_records() if self.records is None else self.records

		if records is not None:
			records.set_step(tick)
		
		prev_mode = model.get_mode()
		
		if self.pbar is not None:
			self.pbar.pause()
		
		self.run_epoch(dataset, model, records)
		
		if self.pbar is not None:
			self.pbar.unpause()
		
		dataset.switch_to(prev_mode)
		model.switch_to(prev_mode)
		if records is not None:
			records.switch_to(prev_mode)
	
@fig.Component('run/checkpoint')
class Checkpointer(RunEvent):
	def __init__(self, A, **kwargs):
		
		path = A.pull('save-root', '<>root', '<>path', None)
		
		limit = A.pull('keep-only', '<>limit', None)
		
		# metric = A.pull('track-best', None)
		
		dataset = A.pull('dataset', None, ref=True)
		model = A.pull('model', None, ref=True)
		records = A.pull('records', None, ref=True)
		clock = A.pull('clock', None, ref=True)
		
		super().__init__(A, **kwargs)
		
		self.root = Path(path)
		self.limit = limit
		if self.limit is not None:
			raise NotImplementedError
		# self.best_metric = metric
		# if self.best_metric is not None:
		# 	raise NotImplementedError
	
		self.dataset = dataset
		self.model = model
		self.records = records
		self.clock = clock
	
	def purge(self):
		self.dataset = None
		self.model = None
		self.records = None
		self.clock = None
	
	def _limit_checkpoints(self, root):
		if self.limit is not None:
			raise NotImplementedError # TODO
	
	def _save_best(self):
		if self.best_metric is not None:
			raise NotImplementedError # TODO
	
	@staticmethod
	def save_checkpoint(path, dataset=None, model=None, records=None, clock=None):

		path = Path(path)
		path.mkdir(exist_ok=True)

		if dataset is not None:
			dataset.checkpoint(path)
		if model is not None:
			model.checkpoint(path)
		if records is not None:
			records.checkpoint(path)
		if clock is not None:
			clock.checkpoint(path)
	
	def create_path(self, num, root=None):
		if root is None:
			root = self.root
	
		if root is None:
			pass
			
		path = Path(root) / f'ckpt{num}'
		return path
	
	def checkpoint(self, dataset=None, model=None, records=None, clock=None, root=None, num=None):
		
		path = self.create_path(num=num, root=root)
		if path is None:
			return
		path.mkdir(exist_ok=True)
		
		if dataset is None:
			dataset = self.dataset
		if model is None:
			model = self.model
		if records is None:
			records = self.records
		if clock is None:
			clock = self.clock
		
		if records is not None:
			records.prep_checkpoint(num)
		
		return self.save_checkpoint(path, dataset, model, records, clock)
		
	def activate(self, tick, info=None):
		assert info is not None
		
		root = info.get_path() if self.root is None else self.root
		
		if root is None:
			return
		
		dataset = info.get_dataset() if self.dataset is None else self.dataset
		model = info.get_model() if self.model is None else self.model
		records = info.get_records() if self.records is None else self.records
		clock = info.get_clock() if self.clock is None else self.clock
		
		self.checkpoint(dataset=dataset, model=model, records=records, clock=clock, root=root, num=tick)
		
		self._limit_checkpoints(root)
		
		
@fig.Component('run/viz')
class VizStep(RunEvent):
	def __init__(self, A, **kwargs):
		
		model = A.pull('model', None, ref=True)
		records = A.pull('records', None, ref=True)
		
		super().__init__(A, **kwargs)
		
		self.model = model
		self.records = records
		
	def activate(self, tick, info=None):
		
		assert info is not None, 'no info provided'
		
		model = info.get_model() if self.model is None else self.model
		records = info.get_records() if self.records is None else self.records
		
		out = info.get_output()
	
		if isinstance(model, Visualizable):
			if records is not None:
				records.set_step(tick)
			model.visualize(out, records)


@fig.Component('run/print')
class PrintStep(RunEvent):
	def activate(self, tick, info=None):
		if info is not None:
			desc = info.get_description(tick)
			print(desc)
			sys.stdout.flush()


@fig.Script('train', description='Train new/existing models')
def iterative_training(A=None, run=None):
	'''
	This is the entry for the training script for new or existing runs.
	Existing runs (or models) can be specified using "path", "load",
	or "resume" with the run name
	'''
	
	respect_config(A)

	if run is None:
		assert A is not None, 'either run or A must not be None'
		run = fig.run('load-run', A)
	
	A = run.get_config()
	
	run.take_steps(complete=True)
	
	include_eval = A.pull('include-eval', True)
	if include_eval:
		evaluate(A=None, run=run)
	
	print(f'Completed {run.get_name()}')
	
	return run






from omnibelt import get_printer
import omnifig as fig

prt = get_printer(__name__)


from ..data import dataset_registry, Downloadable

class DownloadError(Exception):
	pass

@fig.Script('download-dataset', description='Download and format a dataset')
def download_dataset(A, **kwargs):
	'''
	Download and format any registered dataset that subclasses `Downloadable` given the registered name.
	'''
	
	name = A.pull('_dataset_type', '<>dataset-name', '<>name')
	
	cmpn_name = dataset_registry.get(name, None)
	if cmpn_name is None:
		raise DownloadError(f'Can\'t find dataset {name} (has it been registered?)')
	
	cmpn = fig.find_component(cmpn_name).fn
	if not issubclass(cmpn, Downloadable):
		raise DownloadError(f'{name} is not downloadable (it does not subclass `Downloadable`)')

	return cmpn.download(A, **kwargs)






import sys, os
from pathlib import Path
import omnifig as fig
from omnibelt import unspecified_argument

from .. import util

from .loading import respect_config

@fig.Script('eval', description='Evaluate an existing model')
def evaluate(A=None, run=None):
	'''
	Load and evaluate a model (by defaulting using the validation set)
	
	Use argument "use_testset" to evaluate on test set
	'''
	
	if A is not None:
		respect_config(A)
		
	ret_run = False
	if run is None:
		ret_run = True
		assert A is not None, 'either run or A must not be None'
		run = fig.run('load-run', A)
	
	# A = run.get_config()
	
	results = run.evaluate(config=A)
	
	ret_run = A.pull('ret_run', ret_run, silent=True)
	if ret_run:
		return results, run
	return results








import sys, os

import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

import omnifig as fig

from ... import util
from . import utils


def load_inception_model(dim=2048, device='cuda'): # 64, 192, 768, 2048

	if not torch.cuda.is_available(): # cuda not found
		print('WARNING: cuda not found')
		device = 'cpu'
	if device == 'cpu':
		print('WARNING: running on cpu - this will take a long time!')

	block_idx = utils.InceptionV3.BLOCK_INDEX_BY_DIM[dim]

	model = utils.InceptionV3([block_idx]).to(device).eval()
	model._dim = dim
	model._device = device
	return model


def apply_inception(samples, inception, include_grad=False):
	N, C, H, W = samples.shape
	if C == 1:
		samples = torch.cat([samples] * 3, dim=1)
	if C > 3:
		samples = samples[:, :3].contiguous()

	if include_grad:
		pred = inception(samples)[0]
	else:
		with torch.no_grad():
			pred = inception(samples)[0]

	if pred.shape[2] != 1 or pred.shape[3] != 1:
		pred = F.adaptive_avg_pool2d(pred, output_size=(1, 1))

	return pred



def compute_inception_stat(generate, inception=None, batch_size=50, n_samples=50000,
                     dim=2048, device='cuda', pbar=None, name=None):
	"""

	:param generate: callable: input N (int) -> (N, 3, H, W) images (as pytorch tensor)
	:param inception: inception model (if None will automatically load)
	:param batch_size: int
	:param n_samples: int
	:param pbar: tqdm-like (optional)
	:param dim: dimension for inception model (only used if inception is not provided)
	:param device: device for inception model (only used if inception is not provided)
	"""

	if inception is not None:
		dim = inception._dim
		device = inception._device

	pred_arr = np.empty((n_samples, dim)) # TODO: refactor to keep incremental statistics to avoid storing the full set

	if inception is None:
		inception = load_inception_model(dim=dim, device=device)
	
	title = f' {name}' if name is not None else ''
	title = f'Computing{title} FID'

	if pbar is not None:
		pbar = pbar(total=n_samples)
		pbar.set_description(title)
	elif name is not None:
		print(title)

	j = 0
	while j < n_samples:
		N = min(batch_size, n_samples - j)

		samples = generate(N)

		pred = apply_inception(samples, inception)

		pred_arr[j:j+N] = pred.cpu().numpy().reshape(N, -1)

		j += N
		if pbar is not None:
			pbar.update(N)

	if pbar is not None:
		pbar.close()


	m = np.mean(pred_arr, axis=0)
	s = np.cov(pred_arr, rowvar=False)

	return m, s


def compute_frechet_distance(m1, s1, m2, s2, eps=1e-6):
	return utils.calculate_frechet_distance(m1,s1,m2,s2, eps=eps)




import random
import torch

from wrapt import ObjectProxy
from omnibelt import Value as ValueWrapper, primitives, InitWall


from omnifig import Configurable

from .math import set_seed

# region Attributes

class Statelike:
	def state_dict(self, *args, **kwargs):
		try:
			return super().state_dict(*args, **kwargs)
		except AttributeError:
			raise NotImplementedError
	
	def load_state_dict(self, data, **kwargs):
		try:
			return super().load_state_dict(data, **kwargs)
		except AttributeError:
			raise NotImplementedError


class StatelikeList(Statelike, list):
	def state_dict(self):
		return [(x.state_dict() if isinstance(x, Statelike) else None) for x in self]
	
	def load_state_dict(self, data):
		for x, y in zip(self, data):
			if x is not None:
				x.load_state_dict(y)


class StatelikeDict(Statelike, dict):
	def state_dict(self):
		return {k: (v.state_dict() if isinstance(v, Statelike) else None) for k, v in self.items()}
	
	def load_state_dict(self, data):
		for k, v in data.items():
			if v is not None:
				self[k].load_state_dict(v)


# endregion

# region Values

class ValueBase(Statelike, ValueWrapper):
	
	# def __init__(self):
	
	def state_dict(self):
		return self.get()
	
	def load_state_dict(self, data):
		self.set(data)
	
	def item(self):  # compat for hyperparams
		return self.get()
	
	def __getstate__(self):
		return self.state_dict()
	
	def __setstate__(self, state):
		return self.load_state_dict(state)
	
	def __reduce__(self, *args, **kwargs):
		return type(self), (self.get(),)


class Value(ValueBase):
	def __init__(self, A, **kwargs):
		val = A.pull('value', '<>initial')
		assert val is None or isinstance(val, primitives), f'{val} should be a primitive'
		super().__init__(val)
		# self.__wrapped__ = val
		self._self_as_int = A.pull('as-int', isinstance(val, int))
		self.set(val)
		super(ObjectProxy, self).__init__(A, **kwargs)
	
	def set(self, val):
		if hasattr(self, '_self_as_int') and self._self_as_int:
			val = int(val)
		return super().set(val)


# endregion

# region Simple Features

class DimensionBase:
	din, dout = None, None
	
	def __init__(self, *args, din=None, dout=None, **kwargs):
		super().__init__(*args, **kwargs)
		if din is None:
			din = self.din
		if dout is None:
			dout = self.dout
		self.din, self.dout = din, dout
	
	def get_dims(self):
		return self.din, self.dout


class DeviceBase:
	def __init__(self, *args, device=None, **kwargs):
		if device is None:
			device = 'cuda' if torch.cuda.is_available() else 'cpu'
		super().__init__(*args, **kwargs)
		self.device = device
	
	def get_device(self):
		return self.device
	
	def to(self, device):
		self.device = device
		try:
			super().to(device)
		except AttributeError:
			pass
	
	def cuda(self, device=None):
		self.to('cuda' if device is None else device)
	
	def cpu(self):
		self.to('cpu')




class Named(Configurable):
	def __init__(self, A, **kwargs):
		ident = A.pull('_ident', '<>__origin_key', None)
		super().__init__(A, **kwargs)
		self.name = ident
		
	def get_name(self):
		return self.name


class Priority(Configurable):
	def __init__(self, A, **kwargs):
		priority = A.pull('priority', None)
		super().__init__(A, **kwargs)
		self.priority = priority
	
	def get_priority(self):
		return self.priority


class Switchable(Configurable):
	def __init__(self, A, mode=None, **kwargs):
		if mode is None:
			mode = A.pull('mode', 'train')
		super().__init__(A, **kwargs)
		self.mode = mode
	
	def switch_to(self, mode):
		self.mode = mode
	
	def get_mode(self):
		return self.mode


class Deviced(Configurable, DeviceBase):
	def __init__(self, A, device=None, **kwargs):
		if device is None:
			device = 'cuda' if torch.cuda.is_available() else 'cpu'
		device = A.pull('device', device)
		super().__init__(A, device=device, **kwargs)
		self.device = device


class Checkpointable(Configurable):
	def checkpoint(self, path, ident=None):
		pass
	
	def load_checkpoint(self, path, ident=None):
		pass


class Dimensions(Configurable, DimensionBase):
	
	def __init__(self, A, din=None, dout=None, **kwargs):
		if din is None:
			din = A.pull('din', self.din)
		if dout is None:
			dout = A.pull('dout', self.dout)
		super().__init__(A, din=din, dout=dout, **kwargs)
		self.din, self.dout = din, dout


class Seed(Configurable):
	def __init__(self, A, seed=None, **kwargs):
		if seed is None:
			seed = A.pull('seed', random.getrandbits(32))
		set_seed(seed)
		super().__init__(A, **kwargs)
		self.seed = seed


# endregion




import sys, os
import time
from pathlib import Path
import random
from datetime import datetime
from omnibelt import get_now, save_yaml, save_json, load_json, load_yaml

import omnifig as fig

from .clock import Freq

from .. import util
from ..util.features import Configurable, Switchable, Checkpointable, Seed

@fig.Component('records')
class Records(Freq, Switchable, Seed, Configurable, dict):
	def __init__(self, A, **kwargs):

		ckpt = A.pull('_load-ckpt', None, silent=True)
		
		super().__init__(A, **kwargs)
		
		self.update(self._init_info(A))
		self.stats = self._init_stats(A)
		self.logger = self._init_logger(A)
		
		self._stats_format = A.pull('stats_format', 'json')
		self._info_format = A.pull('info_format', 'yaml')
		
		self._log_smooths = A.pull('log-smooth', True)
		self._log_model_str = A.pull('log-model-str', True)
		
		if ckpt is not None:
			self.load_checkpoint(ckpt)
		
		self.purge()
		
	def purge(self):
		self._model = None
		self._dataset = None
		
	def prep(self, order, info=None):
		self._model = info.get_model()
		self._dataset = info.get_dataset()
		
		if self.logger is not None and self._log_model_str:
			self.log('text', 'model-str', str(self._model))
			self.log('text', 'model-optim-str', str(self._model.optim))

	
	def _init_info(self, A):
		info = {
			'total_steps': {},
			'total_samples': {},
			'total_epochs': {},
			
			'checkpoint': None,
			
			'history': A.pull('_history', None, silent=False),
			'argv': sys.argv,
			'timestamp': datetime.now(),
		}
		
		return info
	
	def _init_stats(self, A=None):
		if 'stats' not in A:
			A.push('stats._type', 'stats-manager', silent=True, force_root=True, overwrite=False)
		stats = A.pull('stats', ref=True)
		return stats
		
	def _init_logger(self, A=None):
		if 'logger' not in A:
			A.push('logger._type', 'logger', silent=True, overwrite=False)
		logger = A.pull('logger')
		
		self._use_fmt = A.pull('use_log_fmts', True)
		
		return logger
		
	def activate(self, tick, info=None):
		self.step(tick)
		
	def step(self, num=None, fmt=None):
		
		if self.logger is not None:
			if num is not None:
				self.set_step(num)
			if self._use_fmt and fmt is None:
				fmt = '{}/{}'.format('{}', self.get_mode())
			if fmt is not None:
				self.set_fmt(fmt)
			
			display = self.stats.smooths() if self._log_smooths else self.stats.vals()
			
			for name, val in display.items():
				self.log('scalar', name, val)
		
	def set_fmt(self, fmt=None):
		if self.logger is not None:
			self.logger.set_tag_format(fmt)
	
	def get_fmt(self):
		if self.logger is not None:
			return self.logger.get_tag_format()
		
	def set_step(self, tick):
		if self.logger is not None:
			self.logger.set_step(tick)
		if self.stats is not None:
			self.stats.set_step(tick)
		
	def get_step(self):
		if self.logger is not None:
			self.logger.get_step()
		
	def log(self, data_type, tag, *args, global_step=None, **kwargs):
		if self.logger is not None:
			self.logger.add(data_type, tag, *args, global_step=global_step, **kwargs)
		
	def create_stats(self, *names):
		if self.stats is not None:
			self.stats.new(*names)
		
	def switch_to(self, mode='train'):
		super().switch_to(mode)
		if self.stats is not None:
			self.stats.switch_to(mode)
		
	def prep_checkpoint(self, tick):
		self['checkpoint'] = tick
	
	# region File I/O
	
	def _get_load_fn(self, path):
		return load_json if path.suffix == '.json' else load_yaml

	def import_info(self, path):
		path = Path(path)
		if path.is_dir():
			path = path / f'info.{self._info_format}'
		self.update(self._get_load_fn(path)(path))

	def import_stats(self, path):
		path = Path(path)
		if path.is_dir():
			path = path / f'stats.{self._stats_format}'
		self.stats.load(self._get_load_fn(path)(path))

	def import_(self, path, info=True, stats=True):
		if info:
			self.import_info(path)
		if stats:
			self.import_stats(path)

	def _get_save_fn(self, path):
		return save_json if path.suffix == '.json' else save_yaml

	def export_stats(self, path):
		path = Path(path)
		if path.is_dir():
			path = path / f'stats.{self._stats_format}'
		return self._get_save_fn(path)(self.stats.export(), path)

	def export_info(self, path):
		path = Path(path)
		if path.is_dir():
			path = path / f'info.{self._info_format}'
		return self._get_save_fn(path)(dict(self.items()), path)

	def export(self, path, info=True, stats=True):
		if info:
			self.export_info(path)
		if stats:
			self.export_stats(path)
	
	def checkpoint(self, path, ident=None):
		self.export(path)
	
	def load_checkpoint(self, path, ident=None):
		self.import_(path)


# class Checkpointed(Checkpointable, SimpleRecords):
#
# 	def __init__(self, A, **kwargs):
#
# 		ckpt = A.pull('_load-ckpt', None, silent=True)
#
# 		super().__init__(A, **kwargs)
#
# 		if ckpt is not None:
# 			self.load_checkpoint(ckpt)
#
# 	def checkpoint(self, path, ident=None):
# 		self.export(path)
#
# 	def load_checkpoint(self, path, ident=None):
# 		self.import_(path)
#
#
# @fig.Component('records')
# class Records(Checkpointed, SimpleRecords):
# 	pass






import sys, os
from pathlib import Path

from tqdm import tqdm
import h5py as hf

import omnifig as fig

import xml.etree.ElementTree as ET

from ... import util
# from ..data import Dataset
#
# from ...data import standard_split, Device_Dataset, Info_Dataset, Splitable_Dataset, Testable_Dataset, Batchable_Dataset, Image_Dataset


@fig.Script('format-iam', description='Format IAM handwriting dataset')
def format_iam(A):
	
	root = A.pull('root', '<>path', None)
	if root is None:
		root = util.get_data_dir(A)
	else:
		root = Path(root)
	root = root / 'iam'
	
	lroot = root / 'lines'
	assert lroot.is_dir(), 'Missing IAM Lines dataset (download: https://fki.tic.heia-fr.ch/DBs/iamDB/data/lines.tgz)'
	
	xroot = root / 'xml'
	assert xroot.is_dir(), 'Missing IAM XML meta data (download: https://fki.tic.heia-fr.ch/DBs/iamDB/data/xml.tgz)'
	
	dpath = root / 'lines.h5'
	
	f = hf.File(dpath, 'a+')
	
	if 'text' not in f:
		
		xpaths = list(xroot.glob('**/*.xml'))
		
		
		
		for xpath in tqdm(xpaths):
			tree = ET.parse(xpath)
			root = tree.getroot()
			
			elms = root[1]
			ids = [line.attrib['IDs'] for line in elms]
			lines = [line.attrib['text'] for line in elms]
			
			raise NotImplementedError
		
		pass
	
	
	
	f.close()
	
	pass





# from IPython.core.interactiveshell import InteractiveShell
# InteractiveShell.ast_node_interactivity = "all"
import torch
from torch.utils.data import TensorDataset
# %matplotlib tk
#plt.switch_backend('Qt5Agg') #('Qt5Agg')
from .. import util
from .. import data as datautils

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from matplotlib import animation



def viz_latent(Q, figax=None, figsize=(9, 3), lim_y=None):
	Xs = np.arange(Q.shape[-1]) + 1
	inds = np.stack([Xs] * Q.shape[0])

	vals = Q.detach().cpu().numpy()
	df = pd.DataFrame({'x': inds.reshape(-1), 'y': vals.reshape(-1)})

	if figax is None:
		figax = plt.subplots(figsize=figsize)
	fig, ax = figax

	# plt.figure(fig.num)
	plt.sca(ax)

	hue = None
	split = False
	color = 'C0'
	inner = 'box'
	palette = None

	sns.violinplot(x='x', y='y', hue=hue,
				   data=df, split=split, color=color, palette=palette,
				   scale="count", inner=inner, gridsize=100, )
	if lim_y is not None:
		plt.ylim(-lim_y, lim_y)
	# plt.title('Distributions of Latent Dimensions')
	# plt.xlabel('Dimension')
	# plt.ylabel('Values')
	# plt.tight_layout()
	# border, between = 0.02, 0.01
	# 	plt.subplots_adjust(wspace=between, hspace=between,
	# 						left=border, right=1 - border, bottom=border, top=1 - border)
	return fig, ax

def get_traversal_vecs(Q, steps=32, bounds=None, mnmx=None):

	N, D = Q.shape
	S = steps
	#
	# dH, dW = util.calc_tiling(D)
	#
	# # bounds = (-2,2)
	#
	# save_inds = [0, 1, 2, 3]
	#
	# saved_walks = []

	I = torch.eye(D, device=Q.device).view(1, 1, D, D)

	deltas = torch.linspace(0, 1, steps=S, device=Q.device)
	deltas = torch.stack([deltas] * D)  # DxS

	if mnmx is None:
		if bounds is None:
			mnmx = (Q.min_val(0)[0].view(D, 1), Q.max(0)[0].view(D, 1))
		else:
			mnmx = torch.ones(D,device=Q.device)*bounds[0], torch.ones(D,device=Q.device)*bounds[1]

	mn, mx = mnmx
	mn, mx = mn.view(D, 1), mx.view(D, 1)

	deltas *= mx - mn
	deltas += mn
	deltas = deltas.t().unsqueeze(0).expand(N, S, D).unsqueeze(-1)

	Q = Q.unsqueeze(1).unsqueeze(-1).expand(N, S, D, D)

	vecs = Q * (1 - I) + deltas * I
	vecs = vecs.permute(0, 3, 1, 2) # NxDxSxD (batch, which dim, steps, vec)

	return vecs

def get_traversals(vecs, model, device='cpu', pbar=None): # last dim must be latent dim (model input)

	*shape, D = vecs.shape

	dataset = TensorDataset(vecs.view(-1,D))

	loader = datautils.get_loaders(dataset, device=device, batch_size=64, shuffle=False)

	if pbar is not None:
		loader = pbar(loader)

	imgs = []

	for Q, in loader:
		with torch.no_grad():
			imgs.append(model(Q))

	imgs = torch.cat(imgs)

	_, *img_shape = imgs.shape

	return imgs.view(*shape,*img_shape)

def tile_dims(full):
	tH, tW = util.calc_tiling(len(full))
	N, S, C, H, W = full.shape
	full = full.view(tH, tW, S, C, H, W)
	full = full.permute(2, 3, 0, 4, 1, 5).contiguous().view(S, C, tH * H, tW * W)
	return full

def get_traversal_anim(frames, vals=None, text_fmt='{:2.2f}', text_size=12, scale=1, fps=20):

	frames = frames.permute(0,2,3,1).cpu().numpy()
	if vals is not None:
		vals = vals.cpu().numpy()

	H, W, C = frames[0].shape
	asp = W/H
	fig = plt.figure(figsize=(asp, 1), dpi=int(H*scale),)

	ax = plt.axes([0, 0, 1, 1], frameon=False)
	ax.get_xaxis().set_visible(False)
	ax.get_yaxis().set_visible(False)
	plt.autoscale(tight=True)

	im = plt.imshow(frames[0])
	# plt.axis('off')
	# plt.tight_layout()
	if vals is not None:
		txt = plt.text(5,text_size*H//64, text_fmt.format(vals[0]), size=text_size)
	pass

	plt.close()

	def init():
		im.set_data(frames[0])
		if vals is not None:
			txt.set_text(text_fmt.format(vals[0]))

	def animate(i):
		im.set_data(frames[i])
		if vals is not None:
			txt.set_text(text_fmt.format(vals[i]))
		return im

	anim = animation.FuncAnimation(fig, animate, init_func=init, frames=len(frames), interval=1000//fps)
	return anim

def compute_diffs(walks, dsteps=10):
	'''
	computes the Linf distance in pixel space between an image and the image where
	a single latent dimension is perturbed by approximately half of the range
	of that latent dim
	'''

	B, D, S, C, H, W = walks.shape

	diffs = (walks[:, :, dsteps:] - walks[:, :, :-dsteps]).abs().max(-3)[0]
	diffs = diffs.view(B, D, (S - dsteps) * H * W).mean(-1)
	return diffs


def viz_interventions(dists, figax=None, figsize=(9, 3), color='C2'):

	vals = dists.cpu().numpy()
	Xs = np.arange(vals.shape[-1]) + 1
	inds = np.stack([Xs] * vals.shape[0])
	df = pd.DataFrame({'x': inds.reshape(-1), 'y': vals.reshape(-1)})
	# df['moment']='log(sigma)'

	hue = None
	split = False
	# color = 'C2'
	inner = 'box'
	palette = None

	if figax is None:
		figax = plt.subplots(figsize=figsize)
	fig, ax = figax
	plt.sca(ax)
	sns.violinplot(x='x', y='y', hue=hue,
				   data=df, split=split, color=color, palette=palette,
				   scale="count", inner=inner, gridsize=100, cut=0)
	# plt.title('Intervention Effect on Image')
	# plt.xlabel('Dimension')
	# plt.ylabel('Effect')
	# plt.tight_layout()


	return fig, ax









import sys, os
from pathlib import Path
import random
import numpy as np
import torch
from torch.utils.data import Dataset as PytorchDataset, DataLoader, RandomSampler
import h5py as hf

from omnibelt import save_yaml, load_yaml

import omnifig as fig

from .. import util
from ..util.features import Checkpointable

from .register import dataset_registry
from .collectors import Shuffle_Dataset, Subset_Dataset
from .loaders import Featured_DataLoader, BatchedDataLoader

from ..op.clock import AlertBase

class SimpleDataManager(util.Seed, util.Dimensions, util.Switchable, util.Deviced):
	def __init__(self, A, mode='train', **kwargs):
		
		name = A.pull('_dataset_type', '<>dataset-name', '<>name')
		mods = A.pull('_dataset_mod', [])
		
		default_mode = A.pull('default_mode', '<>mode', mode)
		A.push('mode', default_mode, overwrite=False, silent=True)

		aliases = A.pull('mode-aliases', {})

		super().__init__(A, mode=mode, **kwargs)

		self._default_mode = default_mode
		
		self.A = A

		cmpn_name = dataset_registry.get(name, None)
		if cmpn_name is None:
			raise NotImplementedError
		self.A.push('_type', cmpn_name, silent=True)
		self.A.push('_mod', mods, silent=True)

		self._mode_aliases = aliases

		self.purge()
			
	def startup(self, A=None):
		mode = self._default_mode
		return self.get_data(mode)

	def purge(self):
		self._active = None
		self._modes = {}
	
	def _create_mode(self, mode):

		# self.A.begin()
		if self.A.contains_nodefault(mode):
			self.A.update(self.A.sub(mode))
		self.A.push('mode', mode, silent=True)
		dataset = self.A.pull_self()
		# self.A.abort()
		
		self._modes[mode] = dataset
		return dataset

	def _find_active(self, mode):
		if mode in self._modes:
			return self._modes[mode]
		elif mode in self._mode_aliases:
			return self._find_active(self._mode_aliases[mode])
		return self._create_mode(mode)

	def switch_to(self, mode):
		if mode is None:
			mode = self._default_mode

		super().switch_to(mode)
		self._active = self._find_active(mode)
	def get_data(self, mode=None):
		if mode is not None:
			self.switch_to(mode)
		return self._active
	
	def __len__(self):
		return len(self.get_data())
	
	def __getitem__(self, item):
		return self.get_data()[item]
	
	def __getattribute__(self, item):
		try:
			return super().__getattribute__(item)
		except AttributeError as e:
			if item != '_active' and self._active is not None:
				try:
					return self._active.__getattribute__(item)
				except AttributeError:
					raise e
			else:
				raise e

@fig.AutoModifier('datamanager/loadable')
class Loadable(SimpleDataManager):
	def __init__(self, A, **kwargs):
		num_workers = A.pull('num_workers', 0)
		batch_size = A.pull('batch_size', 64)
		shuffle = A.pull('shuffle', True)
		drop_last = A.pull('drop_last', False)
		loader_device = A.pull('step_device', '<>device', 'cuda' if torch.cuda.is_available() else 'cpu')
		infinite = A.pull('infinite', False)
		extractor = A.pull('extractor', None)
		allow_batched = A.pull('allow_batched', True)
		
		super().__init__(A, **kwargs)
		
		pin_memory = A.pull('pin_memory', self.device == 'cpu')
		
		self._loader_settings = {
			'num_workers': num_workers,
			'batch_size': batch_size,
			'shuffle': shuffle,
			'drop_last': drop_last,
			'device': loader_device,
			'pin_memory': pin_memory,
		}
		self._allow_batched = allow_batched
		self._infinite_loader = infinite
		self._loader_extractor = extractor
	
	def get_loader(self, mode=None, **kwargs):
		dataset = self.get_data(mode=mode)
		return self.to_loader(dataset, **kwargs)
	
	def get_batch_size(self):
		return self._loader_settings.get('batch_size', None)
	
	def to_loader(self, dataset, infinite=None, extractor=None, **updates):
		settings = self._loader_settings.copy()
		settings.update(updates)
		
		loader_cls = Featured_DataLoader
		# loader_cls = DataLoader
		
		# if self._allow_batched:
		# 	try:
		# 		assert dataset.allow_batched()
		# 	except (AttributeError, AssertionError):
		# 		pass
		# 	else:
		# 		# print('Using batched data loader')
		# 		loader_cls = BatchedDataLoader

		if len(dataset) > 1e9:
			generator = settings.get('generator', None)
			seed = settings.get('seed', None)
			shuffle = settings.get('shuffle', False)
			if shuffle:
				if generator is None and seed is not None:
					generator = torch.Generator()
					generator.manual_seed(seed)
					settings['generator'] = generator
				settings['sampler'] = RandomSampler(dataset, replacement=True, generator=generator)
				settings['shuffle'] = False

		loader = loader_cls(dataset, **settings)
		
		if infinite is None:
			infinite = self._infinite_loader
		if extractor is None:
			extractor = self._loader_extractor
		if infinite:
			return util.make_infinite(loader, extractor=extractor)
		return loader


@fig.AutoModifier('datamanager/splitable')
class Splitable(SimpleDataManager):
	def __init__(self, A, skip_load=None, **kwargs):
		
		split = A.pull('split', {})
		shuffle_split = A.pull('shuffle-split', True)
		split_src = A.pull('split-src', 'train')
		
		super().__init__(A, **kwargs)
		
		self.shuffle_split = shuffle_split
		self.split_info = split
		self.split_src = split_src
		self._split_done = False
		
	def _split_load(self, dataset):
		
		if self.split_src == self.get_mode():
			
			skip_modes = dataset.get_available_modes()
			
			self.split_info = {mode: ratio for mode, ratio in self.split_info.items() if mode not in skip_modes}
			
			if len(self.split_info):
			
				modes, ratios = zip(*self.split_info.items())
				
				splits = self.split(dataset, *ratios, shuffle=self.shuffle_split)
				
				for mode, split in zip(modes, splits):
					self._modes[mode] = split
				self._modes[self.split_src] = splits[-1]
				dataset = splits[-1]
				self.switch_to(self.split_src)
		
		return dataset
		
	def startup(self, A=None):
		dataset = super().startup(A=A)
		if not self._split_done:
			dataset = self._split_load(dataset)
			self._split_done = True
		return dataset
		
	def split(self, dataset, *ratios, shuffle=True):
		
		for r in ratios:
			assert 0 < r < 1, f'{r}'
		
		assert sum(ratios) < 1, 'nothing left for remainder'
		
		if shuffle:
			dataset = Shuffle_Dataset(dataset)
		
		parts = []
		idx = 0
		for r in ratios:
			n = int(np.round(len(dataset) * r))
			if n < 1:
				raise Exception(f'invalid ratio: {r} for dataset len {len(dataset)}')
			part = Subset_Dataset(dataset, torch.arange(idx, idx+n))
			parts.append(part)
			idx += n
		last = Subset_Dataset(dataset, torch.arange(idx, len(dataset)))
		parts.append(last)
		
		return parts

class Active(Loadable, AlertBase):
	def __init__(self, A, **kwargs):

		super().__init__(A, **kwargs)
		self.rng = random.Random()
		self.epoch_seed = A.pull('epoch_seed', '<>seed', self.rng.getrandbits(32))
		
		self._loader = None
	
	def _increment_rng(self, seed):
		self.rng.seed(seed)
		return self.rng.getrandbits(32)
	
	def _start_epoch(self, seed=None, **loader_args):
		if seed is None:
			seed = self.epoch_seed
		loader = self.get_loader(seed=seed, **loader_args)
		assert len(loader)
		
		loader = iter(loader)
		self.epoch_seed = self._increment_rng(self.epoch_seed)
		
		return loader
	
	def _end_epoch(self):
		pass
	
	def __next__(self):
		return self.get_batch()
	
	def get_batch(self, force=True, **loader_args):
		
		if self._loader is None:
			self._loader = self._start_epoch(**loader_args)
		
		try:
			return next(self._loader)
		except StopIteration:
			self._end_epoch()
			if force:
				self._loader = None
				return self.get_batch(**loader_args)
			raise
	
	def activate(self, tick, info=None):
		info.receive_batch(self.get_batch())

class InfoManager(Checkpointable, Active):
	def __init__(self, A, **kwargs):
		
		ckpt = A.pull('_load-ckpt', None, silent=True)
		
		super().__init__(A, **kwargs)
		
		self.info = self._init_info(A)
		
		if ckpt is not None:
			self.load_checkpoint(ckpt)
			
		self.purge()
		
	def _init_info(self, A):
		
		info = {
			'batch': None,
			'epoch': None,
			
			'epoch_seed': self.epoch_seed,
		}
		
		return info
	
	def purge(self):
		super().purge()
		self._records = None
	
	def prep(self, order, info=None):
		super().prep(order, info=info)
		
		self._records = info.get_records()

	def _increment_rng(self, seed):
		epoch_seed = super()._increment_rng(seed)
		self.info['epoch_seed'] = epoch_seed
		return epoch_seed
	
	def _start_epoch(self, **loader_args):
		if self.info['batch'] is None:
			self.info['batch'] = 0
		if self.info['epoch'] is None:
			self.info['epoch'] = 0
		loader = super()._start_epoch(**loader_args)
		# loader.skip(self.info['batch'])
		return loader
	
	def _end_epoch(self):
		self.info['batch'] = 0
		self.info['epoch'] += 1
		
		if self._records is not None:
			self._records['total_epochs'][self.get_mode()] += 1
		
		super()._end_epoch()
	
	def get_batch(self, force=True, **loader_args):
		batch = super().get_batch(force=force, **loader_args)
		self.info['batch'] += 1
		# mode = self.get_mode()
		# self._records['total_steps'][mode] += 1
		# self._records['total_samples'][mode] += batch.size(0)
		return batch
	
	def checkpoint(self, path, ident='dataset'):
		path = Path(path)
		
		path = path / f'{ident}.yaml'
		save_yaml(self.info, str(path))
	
	def load_checkpoint(self, path, ident='dataset'):
		path = Path(path)
		
		path = path / f'{ident}.yaml'
		self.info = load_yaml(str(path))
		self.epoch_seed = self.info.get('epoch_seed', self.epoch_seed)
		


@fig.Script('load-data', description='Load datasets')
def load_data(A):

	info = A.pull('dataset', None, raw=True)
	if info is None:
		A.push('_type', 'dataset', silent=True)
	else:
		A = info

	return A.pull_self()


@fig.Component('dataset')
class DataManager(InfoManager, Splitable, SimpleDataManager):
	def __init__(self, A, **kwargs):
		super().__init__(A, **kwargs)
		
		skip_load = A.pull('skip_load', False)
		if not skip_load:
			dataset = self.startup()
			if isinstance(dataset, util.Dimensions):
				self.din, self.dout = dataset.get_dims()
				if A is None:
					A = self.A
				A.push('din', self.din, silent=True)
				A.push('dout', self.dout, silent=True)













import sys, os
from pathlib import Path
import subprocess
import pickle
import zipfile
import h5py as hf
import numpy as np
import torch
from omnibelt import unspecified_argument, get_printer
from torch.nn import functional as F

prt = get_printer(__name__)

try:
	from google.cloud import storage
except ImportError:
	prt.info('Failed to import gsutil (install with "pip install gsutil")')

try:
	import wget
except ImportError:
	prt.info('Failed to import wget (install with "pip install wget")')
	

from ... import util
from ...data import Dataset, Deviced, Batchable, Image_Dataset, Downloadable, DatasetBase, MissingDatasetError

from .transforms import Cropped

def _rec_decode(obj):
	'''
	recursively convert bytes to str
	:param obj: root obj
	:return:
	'''
	if isinstance(obj, dict):
		return {_rec_decode(k):_rec_decode(v) for k,v in obj.items()}
	if isinstance(obj, list):
		return [_rec_decode(x) for x in obj]
	if isinstance(obj, tuple):
		return tuple(_rec_decode(x) for x in obj)
	if isinstance(obj, bytes):
		return obj.decode()
	return obj

@Dataset('dsprites')
class dSprites(Deviced, Batchable, Image_Dataset):

	din = (1, 64, 64)
	dout = 5

	def __init__(self, A):

		dataroot = A.pull('dataroot', None)
		root = None

		label_type = A.pull('label_type', None)

		din = A.pull('din', self.din)
		dout = A.pull('dout', din if label_type is None else self.dout)

		filename = A.pull('filename', 'dsprites_ndarray_co1sh3sc6or40x32y32_64x64.npz')

		assert label_type in {None, 'value', 'class'}, 'Unknown type of label: {}'.format(label_type)
		# assert images is not None or dataroot is not None, 'nothing to use/load'

		if dout is None and label_type is not None:
			dout = 5 if label_type == 'value' else 113

		super().__init__(din=din, dout=din if dout is None else dout)

		if dataroot is not None:
			root = os.path.join(dataroot, 'dsprites')
			path = os.path.join(root, filename)
			print('Loading dSprites dataset from disk: {}'.format(path))
			data = np.load(path, allow_pickle=True, encoding='bytes')

			self.meta = _rec_decode(data['metadata'][()])

			images = torch.from_numpy(data['imgs']).unsqueeze(1)

			self.register_buffer('images', images)

			if label_type is not None:
				if label_type == 'value':
					labels = torch.from_numpy(data['latents_values'][:,1:]).float()
				else:
					labels = torch.from_numpy(data['latents_classes'][:,1:]).int()
				self.register_buffer('labels', labels)

		self.labeled = hasattr(self, 'labels')
		self.root = root

	def get_raw_data(self):
		if self.labeled:
			return self.images, self.labels
		return self.images

	def __len__(self):
		return len(self.images)

	def __getitem__(self, item):
		imgs = self.images[item].float()
		if self.labeled:
			return imgs, self.labels[item]
		return imgs,


@Dataset('3dshapes')
class Shapes3D(Deviced, Batchable, Downloadable, Image_Dataset):

	din = (3, 64, 64)
	dout = 6

	def __init__(self, A, mode=None, labeled=None, **kwargs):

		if mode is None:
			mode = A.pull('mode', 'full')
		if labeled is None:
			labeled = A.pull('labeled', False)
		label_type = A.pull('label_type', 'class') if labeled else None

		din = A.pull('din', self.din)
		dout = A.pull('dout', self.dout if labeled else din)

		super().__init__(A, din=din, dout=dout, **kwargs)
		self.add_existing_modes('test')
		
		self.factor_order = ['floor_hue', 'wall_hue', 'object_hue', 'scale', 'shape', 'orientation']
		self.factor_sizes = [10, 10, 10, 8, 4, 15]
		self.factor_num_values = {'floor_hue': 10, 'wall_hue': 10, 'object_hue': 10,
		                          'scale': 8, 'shape': 4, 'orientation': 15}
		
		raw_mins = torch.tensor([  0.  ,   0.  ,   0.  ,   0.75,   0.  , -30.  ]).float()
		raw_maxs = torch.tensor([  0.9 ,  0.9 ,  0.9 ,  1.25,  3.  , 30.  ]).float()
		
		dataroot = self.root / '3dshapes'
		dataroot.mkdir(exist_ok=True)
		self.root = dataroot
		
		path = dataroot / '3dshapes.h5'
		if not path.is_file():
			download = A.pull('download', False)
			if download:
				self.download(A)
			else:
				raise MissingDatasetError('3dshapes')
		
		self.images, self.labels, self.indices = None, None, None
		with hf.File(path, 'r') as data:
			
			images = torch.from_numpy(data['images'][()])
			labels = torch.from_numpy(data['labels'][()]).float() if labeled else None
			
			indices = None if mode == 'full' else \
				(data['test_idx'][()] if mode == 'test' else data['train_idx'][()])
			if indices is not None:
				indices = torch.from_numpy(indices).long()
				images = images[indices]
				if labels is not None:
					labels = labels[indices]
				self.register_buffer('indices', indices)
			else:
				prt.warning('using the full dataset (train+test)')
			
			images = images.permute(0, 3, 1, 2)#.float().div(255)
			self.register_buffer('images', images)
			
			if labels is not None:
				if label_type == 'class':
					labels -= raw_mins
					labels *= ((torch.tensor(self.factor_sizes).float() - 1) / raw_maxs)
					labels = labels.long()
				self.register_buffer('labels', labels)
				
		self.labeled = labeled
		self.label_type = label_type

		
	_source_url = 'gs://3d-shapes/3dshapes.h5'
	@classmethod
	def download(cls, A, dataroot=None, **kwargs):
		
		if dataroot is None:
			dataroot = util.get_data_dir(A) / '3dshapes'
		
		dest = dataroot / '3dshapes.h5'
		
		force_download = A.pull('force-download', False, silent=True)
		
		if not dest.is_file() or force_download:
			print(f'Downloading 3dshapes dataset to {str(dest)} ...', end='')
			subprocess.run(['gsutil', 'cp', cls._source_url, str(dest)])
			print(' done!')
		
		ratio = A.pull('separate-testset', 0.2)
		
		rng = np.random.RandomState(0)
		
		with hf.File(dest, 'r+') as f:
			if ratio is not None and ratio > 0 and 'train_idx' not in f:
				
				N, H, W, C = f['images'].shape
				
				test_N = int(N * ratio)
				
				order = rng.permutation(N)
				
				train_idx = order[:-test_N]
				train_idx.sort()
				test_idx = order[-test_N:]
				test_idx.sort()
				
				f.create_dataset('train_idx', data=train_idx)
				f.create_dataset('test_idx', data=test_idx)

	def get_factor_sizes(self):
		return self.factor_sizes
	def get_factor_order(self):
		return self.factor_order

	def get_labels(self):
		return self.labels

	def get_raw_data(self):
		if self.labeled:
			return self.images, self.labels
		return self.images

	def __len__(self):
		return len(self.images)

	def __getitem__(self, item):
		# if isinstance(item, (np.ndarray, torch.Tensor)):
		# 	item = item.tolist()
		images = self.images[item].float().div(255)
		# if self.noise is not None:
		# 	images = images.add(torch.randn_like(images).mul(self.noise)).clamp(0,1)
		if self.labeled:
			labels = self.labels[item]
			# labels = torch.from_numpy(self.labels[item]).float()
			return images, labels
		return images,


@Dataset('full-celeba')  # probably shouldnt be used
class FullCelebA(Downloadable, Image_Dataset):  # TODO: automate downloading and formatting
	
	din = (3, 218, 178)
	
	ATTRIBUTES = ['5_o_Clock_Shadow', 'Arched_Eyebrows', 'Attractive', 'Bags_Under_Eyes',
	              'Bald', 'Bangs', 'Big_Lips', 'Big_Nose', 'Black_Hair', 'Blond_Hair',
	              'Blurry', 'Brown_Hair', 'Bushy_Eyebrows', 'Chubby', 'Double_Chin',
	              'Eyeglasses', 'Goatee', 'Gray_Hair', 'Heavy_Makeup', 'High_Cheekbones',
	              'Male', 'Mouth_Slightly_Open', 'Mustache', 'Narrow_Eyes', 'No_Beard',
	              'Oval_Face', 'Pale_Skin', 'Pointy_Nose', 'Receding_Hairline', 'Rosy_Cheeks',
	              'Sideburns', 'Smiling', 'Straight_Hair', 'Wavy_Hair', 'Wearing_Earrings',
	              'Wearing_Hat', 'Wearing_Lipstick', 'Wearing_Necklace', 'Wearing_Necktie', 'Young', ]
	
	def __init__(self, A, resize=unspecified_argument, label_type=unspecified_argument, mode=None, **kwargs):
		
		if label_type is unspecified_argument:
			label_type = A.pull('label_type', None)
		
		if mode is None:
			mode = A.pull('mode', 'train')
		if resize is unspecified_argument:
			resize = A.pull('resize', (256, 256))
		
		din = A.pull('din', self.din)
		
		if label_type is None:
			dout = din
		elif label_type == 'attr':
			dout = 40
		elif label_type == 'landmark':
			dout = 10
		elif label_type == 'identity':
			dout = 1
		else:
			raise Exception('unknown {}'.format(label_type))
		
		dout = A.pull('dout', dout)
		
		_labels = {
			'attr': 'attrs',
			'identity': 'identities',
			'landmark': 'landmarks',
		}
		
		if resize is not None:  # TODO: use Interpolated as modifier
			din = 3, *resize
		
		super().__init__(A, din=din, dout=dout, **kwargs)
		# self.add_existing_modes('val', 'test')
		self.root = Path(self.root) / 'celeba'
		dataroot = self.root
		name = 'celeba_test.h5' if mode == 'test' else 'celeba_train.h5'
		
		with hf.File(dataroot/name, 'r') as f:
			self.images = f['images'][()]  # encoded as str
			self.labels = f[_labels[label_type]][()] if label_type is not None else None
			
			self.attr_names = f.attrs['attr_names']
			self.landmark_names = f.attrs['landmark_names']
		
		self.resize = resize
	
	# google drive ids
	_google_drive_ids = {
		'list_eval_partition.txt': '0B7EVK8r0v71pY0NSMzRuSXJEVkk',
		'identity_CelebA.txt': '1_ee_0u7vcNLOfNLegJRHmolfH5ICW-XS',
		'list_attr_celeba.txt': '0B7EVK8r0v71pblRyaVFSWGxPY0U',
		'list_bbox_celeba.txt': '0B7EVK8r0v71pbThiMVRxWXZ4dU0',
		'list_landmarks_align_celeba.txt': '0B7EVK8r0v71pd0FJY3Blby1HUTQ',
	}
	_google_drive_image_id = '0B7EVK8r0v71pZjFTYXZWM3FlRnM'
	
	@classmethod
	def download(cls, A, dataroot=None, **kwargs):
		
		raise NotImplementedError('Unfortunately, automatically downloading CelebA is current not supported.')
		
		dataroot = util.get_data_dir(A)
		dataroot = dataroot / 'celeba'
		dataroot.mkdir(exist_ok=True)
		
		prt.warning(
			'Downloading CelebA doesn\'t seem to work currently (issues with google drive), instead it must be downloaded manually:'
			'\n1. create a directory "celeba" in the local_data directory'
			'\n2. download "img_align_celeba.zip" into that directory'
			'\n3. download label files: {}'
			'\n4. extract "img_align_celeba.zip" to a directory "img_align_celeba"'.format(
				', '.join(f'"{c}"' for c in cls._google_drive_ids.keys())))
		
		raise NotImplementedError
		
		imgdir = dataroot / 'img_align_celeba'
		
		pbar = A.pull('pbar', None)
		
		util.download_file_from_google_drive('0B7EVK8r0v71pOXBhSUdJWU1MYUk', dataroot, pbar=pbar)
		
		for name, ID in cls._google_drive_ids.items():
			path = dataroot / name
			if not path.exists():
				util.download_file_from_google_drive(ID, dataroot, name, pbar=pbar)
		
		if not imgdir.is_dir():
			
			imgpath = dataroot / 'img_align_celeba.zip'
			
			# download zip
			if not imgpath.exists():
				imgpath = util.download_file_from_google_drive(cls._google_drive_image_id, dataroot, pbar=pbar)
			
			# extract zip
			imgdir.mkdir(exist_ok=True)
			with zipfile.ZipFile(str(imgpath), 'r') as zip_ref:
				zip_ref.extractall(str(imgdir))
			
			# os.remove(str(imgpath))
			
		raise NotImplementedError
	
	def get_factor_sizes(self):
		return [2] * len(self.ATTRIBUTES)
	
	def get_factor_order(self):
		return self.ATTRIBUTES
	
	def get_attribute_key(self, idx):
		try:
			return self.ATTRIBUTES[idx]
		except IndexError:
			pass
	
	def __len__(self):
		return len(self.images)
	
	def __getitem__(self, item):
		
		img = torch.from_numpy(util.str_to_jpeg(self.images[item])).permute(2, 0, 1).float().div(255)
		
		if self.resize is not None:
			img = F.interpolate(img.unsqueeze(0), size=self.resize, mode='bilinear').squeeze(0)
		
		if self.labels is None:
			return img,
		
		lbl = torch.from_numpy(self.labels[item])
		
		return img, lbl


@Dataset('celeba')
class CelebA(Cropped, FullCelebA):
	def __init__(self, A, resize=None, crop_size=128, **kwargs):
		# raise NotImplementedError('doesnt work since FullCelebA automatically resizes to (256,256)')
		crop_size = A.pull('crop_size', crop_size) # essentially sets this as the default

		super().__init__(A, resize=resize, crop_size=crop_size, **kwargs)


@Dataset('mpi3d')
class MPI3D(Deviced, Batchable, Downloadable, Image_Dataset):

	din = (3, 64, 64)
	dout = 7

	def __init__(self, A, mode=None, fid_ident=None, **kwargs):

		if mode is None:
			mode = A.pull('mode', 'train')
		labeled = A.pull('labeled', False)

		din = A.pull('din', self.din)
		dout = A.pull('dout', self.dout if labeled else din)

		cat = A.pull('category', 'toy')

		assert cat in {'toy', 'sim', 'realistic', 'real', 'complex'}, 'invalid category: {}'.format(cat)
		if cat == 'sim':
			cat = 'realistic'

		super().__init__(A, din=din, dout=dout, fid_ident=cat, **kwargs)
		self.add_existing_modes('test')
		dataroot = self.root / 'mpi3d'
		self.root = dataroot
		
		self.factor_order = ['object_color', 'object_shape', 'object_size', 'camera_height', 'background_color',
		                     'horizonal_axis', 'vertical_axis']
		self.factor_sizes = [6,6,2,3,3,40,40]

		self.factor_values = {
			'object_color': ['white', 'green', 'red', 'blue', 'brown', 'olive'],
			'object_shape': ['cone', 'cube', 'cylinder', 'hexagonal', 'pyramid', 'sphere'],
			'object_size': ['small', 'large'],
			'camera_height': ['top', 'center', 'bottom'],
			'background_color': ['purple', 'sea_green', 'salmon'],
			'horizonal_axis': list(range(40)),
			'vertical_axis': list(range(40)),
		}
		
		if cat == 'complex':
		
			self.factor_sizes[0], self.factor_sizes[1] = 4, 4
			self.factor_values['object_shape'] = ['mug', 'ball', 'banana', 'cup']
			self.factor_values['object_color'] = ['yellow', 'green', 'olive', 'red']

		sizes = np.array(self.factor_sizes)

		flr = np.cumprod(sizes[::-1])[::-1]
		flr[:-1] = flr[1:]
		flr[-1] = 1

		self._sizes = torch.from_numpy(sizes.copy()).long()
		self._flr = torch.from_numpy(flr.copy()).long()

		self.labeled = labeled

		path = dataroot / f'mpi3d_{cat}.h5'
		
		if not path.is_file():
			download = A.pull('download', False)
			if download:
				self.download(A)
			else:
				raise MissingDatasetError(f'mpi3d-{cat}')
		
		with hf.File(path, 'r') as f:
			if mode == 'full':
				images = np.concatenate([f['train_images'][()], f['test_images'][()]])
				indices = np.concatenate([f['train_idx'], f['test_idx'][()]])
			elif mode == 'test':
				images = f['test_images'][()]
				indices = f['test_idx'][()]
			else:
				images = f['train_images'][()]
				indices = f['train_idx'][()]
		
		ordered = A.pull('ordered', mode == 'full')
		if ordered:
			self.register_buffer('sel_index', torch.from_numpy(indices.argsort()))
		else:
			self.sel_index = None
		
		self.register_buffer('images', torch.from_numpy(images).permute(0,3,1,2))
		self.register_buffer('indices', torch.from_numpy(indices))

	
	_source_url = {
		'toy': 'https://storage.googleapis.com/disentanglement_dataset/Final_Dataset/mpi3d_toy.npz',
		'realistic': 'https://storage.googleapis.com/disentanglement_dataset/Final_Dataset/mpi3d_realistic.npz',
		'real': 'https://storage.googleapis.com/disentanglement_dataset/Final_Dataset/mpi3d_real.npz',
	}
	@classmethod
	def download(cls, A, dataroot=None, cat=None, **kwargs):
		
		if dataroot is None:
			dataroot = util.get_data_dir(A)
			dataroot = dataroot / 'mpi3d'
		dataroot.mkdir(exist_ok=True)
		
		if cat is None:
			cat = A.pull('category', 'toy')
		
		assert cat in {'toy', 'sim', 'realistic', 'real'}, f'invalid category: {cat}'
		if cat == 'sim':
			cat = 'realistic'
		
		ratio = A.pull('separate-testset', 0.2)
		
		path = dataroot / f'mpi3d_{cat}.h5'
		
		if not path.exists():
			rawpath = dataroot / f'mpi3d_{cat}.npz'
			if not rawpath.exists():
				print(f'Downloading mpi3d-{cat}')
				wget.download(cls._source_url[cat], str(rawpath))
			
			print('Loading full dataset into memory to split train/test')
			full = np.load(rawpath)['images']
			
			N = len(full)
		
			test_N = int(N * ratio)

			rng = np.random.RandomState(0)
			order = rng.permutation(N)
			
			train_idx = order[:-test_N]
			train_idx.sort()
			test_idx = order[-test_N:]
			test_idx.sort()
			
			with hf.File(path, 'w') as f:
				f.create_dataset('train_idx', data=train_idx)
				f.create_dataset('train_images', data=full[train_idx])
				
				f.create_dataset('test_idx', data=test_idx)
				f.create_dataset('test_images', data=full[test_idx])
				
			os.remove(str(rawpath))

	def get_factor_sizes(self):
		return self.factor_sizes
	def get_factor_order(self):
		return self.factor_order

	def get_labels(self):
		return self.get_label(self.indices)

	def get_label(self, inds):
		try:
			len(inds)
			inds = inds.view(-1,1)
		except TypeError:
			pass

		lvls = inds // self._flr
		labels = lvls % self._sizes

		return labels

	def __len__(self):
		return len(self.images)

	def __getitem__(self, idx):
		if self.sel_index is not None:
			idx = self.sel_index[idx]
		imgs = self.images[idx].float().div(255)
		if self.labeled:
			labels = self.get_label(self.indices[idx])
			return imgs, labels
		return imgs,


class OldFullCelebA(Downloadable, Image_Dataset):  # TODO: automate downloading and formatting
	
	din = (3, 218, 178)
	
	ATTRIBUTES = ['5_o_Clock_Shadow', 'Arched_Eyebrows', 'Attractive', 'Bags_Under_Eyes',
	              'Bald', 'Bangs', 'Big_Lips', 'Big_Nose', 'Black_Hair', 'Blond_Hair',
	              'Blurry', 'Brown_Hair', 'Bushy_Eyebrows', 'Chubby', 'Double_Chin',
	              'Eyeglasses', 'Goatee', 'Gray_Hair', 'Heavy_Makeup', 'High_Cheekbones',
	              'Male', 'Mouth_Slightly_Open', 'Mustache', 'Narrow_Eyes', 'No_Beard',
	              'Oval_Face', 'Pale_Skin', 'Pointy_Nose', 'Receding_Hairline', 'Rosy_Cheeks',
	              'Sideburns', 'Smiling', 'Straight_Hair', 'Wavy_Hair', 'Wearing_Earrings',
	              'Wearing_Hat', 'Wearing_Lipstick', 'Wearing_Necklace', 'Wearing_Necktie', 'Young', ]
	
	def __init__(self, A, resize=unspecified_argument, label_type=unspecified_argument, mode=None, **kwargs):
		self.add_existing_modes('val', 'test')
		
		if label_type is unspecified_argument:
			label_type = A.pull('label_type', None)
		
		if mode is None:
			mode = A.pull('mode', 'train')
		if resize is unspecified_argument:
			resize = A.pull('resize', (256, 256))
		
		din = A.pull('din', self.din)
		
		if label_type is None:
			dout = din
		elif label_type == 'attr':
			dout = 40
		elif label_type == 'landmark':
			dout = 10
		elif label_type == 'identity':
			dout = 1
		else:
			raise Exception('unknown {}'.format(label_type))
		
		dout = A.pull('dout', dout)
		
		_labels = {
			'attr': 'attrs',
			'identity': 'identities',
			'landmark': 'landmarks',
		}
		
		if resize is not None:  # TODO: use Interpolated as modifier
			din = 3, *resize
		
		super().__init__(A, din=din, dout=dout, **kwargs)
		self.root = Path(self.root) / 'celeba'
		dataroot = self.root
		name = 'celeba_test.h5' if mode == 'test' else 'celeba_train.h5'
		
		with hf.File(os.path.join(dataroot, 'celeba', name), 'r') as f:
			self.images = f['images'][()]  # encoded as str
			self.labels = f[_labels[label_type]][()] if label_type is not None else None
			
			self.attr_names = f.attrs['attr_names']
			self.landmark_names = f.attrs['landmark_names']
		
		self.resize = resize
	
	# google drive ids
	_google_drive_ids = {
		'img_align_celeba.zip': '0B7EVK8r0v71pZjFTYXZWM3FlRnM',
		'list_eval_partition.txt': '0B7EVK8r0v71pY0NSMzRuSXJEVkk',
		'identity_CelebA.txt': '1_ee_0u7vcNLOfNLegJRHmolfH5ICW-XS',
		'list_attr_celeba.txt': '0B7EVK8r0v71pblRyaVFSWGxPY0U',
		'list_bbox_celeba.txt': '0B7EVK8r0v71pbThiMVRxWXZ4dU0',
		'list_landmarks_align_celeba.txt': '0B7EVK8r0v71pd0FJY3Blby1HUTQ',
	}
	
	def download(cls, A, dataroot=None, **kwargs):
		
		dataroot = util.get_data_dir(A)
		dataroot = dataroot / 'celeba'
		
		raise NotImplementedError
	
	def get_factor_sizes(self):
		return [2] * len(self.ATTRIBUTES)
	
	def get_factor_order(self):
		return self.ATTRIBUTES
	
	def get_attribute_key(self, idx):
		try:
			return self.ATTRIBUTES[idx]
		except IndexError:
			pass
	
	def __len__(self):
		return len(self.images)
	
	def __getitem__(self, item):
		
		img = torch.from_numpy(util.str_to_jpeg(self.images[item])).permute(2, 0, 1).float().div(255)
		
		if self.resize is not None:
			img = F.interpolate(img.unsqueeze(0), size=self.resize, mode='bilinear').squeeze(0)
		
		if self.labels is None:
			return img,
		
		lbl = torch.from_numpy(self.labels[item])
		
		return img, lbl







import sys, os, time
import random
import numpy as np
import torch
import torch.multiprocessing as mp

from omnibelt import unspecified_argument
import omnifig as fig

from .collectors import DatasetBase

from .. import util


class SamplerBase:
	def __init__(self):
		self.units = None

	def _set_units(self):
		sz = self.factors_num_values.tolist()
		sz.append(1)
		self.units = torch.from_numpy(np.cumprod(sz[::-1])[-2::-1].copy()).long()

	def __len__(self):
		if self.units is None:
			self._set_units()
		return len(self.units)
	
	def sample_labels(self, batch, random_state=None):
		# if random_state is None:
		# 	random_state = torch
		sizes = self.factors_num_values
		return torch.rand(batch, len(sizes)).mul(sizes).long()
	
	def labels_to_inds(self, labels):
		if isinstance(labels, np.ndarray):
			labels = torch.from_numpy(labels)
		if self.units is None:
			self._set_units()
		return labels.matmul(self.units.unsqueeze(-1)).squeeze()
	
	@property
	def num_factors(self):
		return len(self.factors_num_values)
	
	@property
	def factors_num_values(self):
		raise NotImplementedError
		
	@property
	def observation_shape(self):
		raise NotImplementedError
	
	def inds_to_samples(self, inds):
		return self[inds]


class Intervention_Sampler(SamplerBase):
	def __init__(self, dataset, include_labels=False):
		super().__init__()
		self.dataset = dataset
		self.include_labels = include_labels
	
	@property
	def factors_num_values(self):
		sizes = self.dataset.get_factor_sizes()
		if isinstance(sizes, list):
			sizes = torch.tensor(sizes)
		return sizes
	
	@property
	def observation_shape(self):
		return self.dataset.din
	
	def inds_to_samples(self, inds):
		imgs, lbls = self.dataset[inds]
		if self.include_labels:
			return imgs, lbls
		return imgs
	
	def full_intervention(self, idx=None):
		if idx is None:
			idx = random.randint(0, self.num_factors-1)
		vals = self.factors_num_values[idx]
		sample = self.sample_labels(1).expand(vals, self.num_factors).clone()
		sample[:, idx] = torch.arange(vals)
		inds = self.labels_to_inds(sample)
		return self.inds_to_samples(inds)
	
	def intervention(self, idx=None, B=128):
		sizes = self.factors_num_values
		sample = self.sample_labels(B)
		sample[:, idx] = torch.randint(sizes[idx], size=())
		inds = self.labels_to_inds(sample)
		return self.inds_to_samples(inds)


@fig.Component('intervention-sampler')
class _InterventionSampler(util.Configurable, Intervention_Sampler):
	def __init__(self, A, dataset=None, sizes=unspecified_argument, include_labels=None, **kwargs):
		
		if dataset is None:
			dataset = A.pull('dataset', ref=True)
		
		if sizes is unspecified_argument:
			sizes = A.pull('sizes', None)
		
		if include_labels is None:
			include_labels = A.pull('include-labels', False)
			
		super().__init__(A, dataset=dataset, sizes=sizes, include_labels=include_labels, **kwargs)


@fig.AutoModifier('joint-sampler')
class JointFactorSampler(DatasetBase, util.Configurable, SamplerBase):
	'''
	for datasets of the type: (observation, factors)
	where the observation is usually an image, and the factors are the latent factors of variation that (ideally)
	include all causal variables in the SCM of the underlying process that produced the observations.
	'''
	def __init__(self, A, to_numpy=None, **kwargs):
		
		if to_numpy:
			to_numpy = A.pull('to_numpy', False)
		
		super().__init__(A, **kwargs)
		
		self.to_numpy = to_numpy
	
	@property
	def factors_num_values(self):
		return torch.tensor(self.get_factor_sizes())
	
	@property
	def observation_shape(self):
		return self.din
	
	def inds_to_samples(self, inds):
		out = self[inds]
		if isinstance(out, tuple):
			return out[0]
		return out
	
	def sample_factors(self, num, random_state, to_numpy=None):
		"""Sample a batch of factors Y."""
		if to_numpy is None:
			to_numpy = self.to_numpy
		factors = self.sample_labels(num, random_state=random_state)
		if to_numpy:
			return factors.numpy()
		return factors
	
	def sample_observations_from_factors(self, factors, random_state, to_numpy=None):
		"""Sample a batch of observations X given a batch of factors Y."""
		if to_numpy is None:
			to_numpy = self.to_numpy
		observations = self.inds_to_samples(self.labels_to_inds(factors))
		if to_numpy:
			observations = observations.numpy()
		return observations
	
	def sample(self, num, random_state, to_numpy=None):
		"""Sample a batch of factors Y and observations X."""
		if to_numpy is None:
			to_numpy = self.to_numpy
		factors = self.sample_factors(num, random_state, to_numpy=False)
		observations = self.sample_observations_from_factors(factors, random_state)
		if to_numpy:
			factors = factors.numpy()
			observations = observations.numpy()
		return factors, observations
	
	def sample_observations(self, num, random_state, to_numpy=None):
		"""Sample a batch of observations X."""
		return self.sample(num, random_state, to_numpy=to_numpy)[1]
		









from itertools import chain
import torch
from torch import optim as O
from torch import nn
from torch.optim import Optimizer as PytorchOptimizer

from omnibelt import LoadedValue, get_printer

prt = get_printer(__name__)

from .stats import StatsMeter

import omnifig as fig

# region Optimizers

class OptimizerBase(PytorchOptimizer):
	def __init__(self, A=None, **settings):
		super().__init__(params=[torch.zeros(0)], **settings)
		self.param_groups.clear()
	
	def __setstate__(self, state):
		groups = state.get('param_groups', None)
		if groups is not None:
			for group, new in zip(self.param_groups, groups):
				group.update({k:v for k,v in new.items() if not isinstance(v, LoadedValue)})
			del state['param_groups']
		super().__setstate__(state)
	
	def prep(self, params):
		
		param_groups = list(params)
		if len(param_groups) == 0:
			raise ValueError("optimizer got an empty parameter list")
		if not isinstance(param_groups[0], dict):
			param_groups = [{'params': param_groups}]
		
		for param_group in param_groups:
			self.add_param_group(param_group)


@fig.AutoComponent('sgd', auto_name=False)
class SGD(OptimizerBase, O.SGD):
	def __init__(self, lr, momentum=0, dampening=0, weight_decay=0, nesterov=False):
		super().__init__(lr=lr, momentum=momentum, dampening=dampening, weight_decay=weight_decay, nesterov=nesterov)

@fig.AutoComponent('asgd', auto_name=False)
class ASGD(OptimizerBase, O.ASGD):
	def __init__(self, lr=0.01, lambd=0.0001, alpha=0.75, t0=1000000.0, weight_decay=0):
		super().__init__(lr=lr, lambd=lambd, alpha=alpha, t0=t0, weight_decay=weight_decay)


@fig.AutoComponent('adadelta', auto_name=False)
class Adagrad(OptimizerBase, O.Adadelta):
	def __init__(self, lr=1.0, rho=0.9, eps=1e-06, weight_decay=0):
		super().__init__(lr=lr, rho=rho, weight_decay=weight_decay, eps=eps)
		
@fig.AutoComponent('adagrad', auto_name=False)
class Adagrad(OptimizerBase, O.Adagrad):
	def __init__(self, lr=0.01, lr_decay=0, weight_decay=0, initial_accumulator_value=0, eps=1e-10):
		super().__init__(lr=lr, lr_decay=lr_decay, weight_decay=weight_decay,
		                 initial_accumulator_value=initial_accumulator_value, eps=eps)


@fig.AutoComponent('adam', auto_name=False)
class Adam(OptimizerBase, O.Adam):
	def __init__(self, lr=0.001, beta1=0.9, beta2=0.999, eps=1e-8, weight_decay=0, amsgrad=False):
		super().__init__(lr=lr, betas=(beta1, beta2), eps=eps, weight_decay=weight_decay, amsgrad=amsgrad)

@fig.AutoComponent('adamw', auto_name=False)
class AdamW(OptimizerBase, O.AdamW):
	def __init__(self, lr=0.001, beta1=0.9, beta2=0.999, eps=1e-8, weight_decay=1e-2, amsgrad=False):
		super().__init__(lr=lr, betas=(beta1, beta2), eps=eps, weight_decay=weight_decay, amsgrad=amsgrad)

@fig.AutoComponent('adamax', auto_name=False)
class Adamax(OptimizerBase, O.Adamax):
	def __init__(self, lr=0.002, beta1=0.9, beta2=0.999, eps=1e-08, weight_decay=0):
		super().__init__(lr=lr, betas=(beta1, beta2), eps=eps, weight_decay=weight_decay)


@fig.AutoComponent('rmsprop', auto_name=False)
class RMSprop(OptimizerBase, O.RMSprop):
	def __init__(self, lr=0.01, alpha=0.99, eps=1e-08, weight_decay=0, momentum=0, centered=False):
		super().__init__(lr=lr, alpha=alpha, eps=eps, weight_decay=weight_decay, momentum=momentum, centered=centered)

@fig.AutoComponent('rprop', auto_name=False)
class Rprop(OptimizerBase, O.Rprop):
	def __init__(self, lr=0.01, eta1=0.5, eta2=1.2, step1=1e-06, step2=50):
		super().__init__(lr=lr, etas=(eta1, eta2), step_sizes=(step1, step2))


try:
	from ranger import Ranger as RangerOptim
	
	@fig.AutoComponent('ranger', auto_name=False)
	class Ranger(OptimizerBase, RangerOptim):
		def __init__(self, lr=0.001, alpha=0.5, k=6, N_sma_threshhold=5, beta1=.95, beta2=0.999, eps=1e-5, weight_decay=0):
			super().__init__(lr=lr, alpha=alpha, k=k, N_sma_threshhold=N_sma_threshhold, betas=(beta1,beta2),
			                 eps=eps, weight_decay=weight_decay)

except ImportError:
	prt.info('failed to import Ranger optimizer')



# endregion


# @fig.AutoModifier('schedulable')
# class Schedulable(BaseOptimizer):
# 	def __init__(self, A):
# 		scheduler = A.pull('scheduler', None)
# 		super().__init__(A)
#
# 		self.scheduler = scheduler
#
# 	def prep(self, params):
# 		super().prep(params)
# 		if self.scheduler is not None:
# 			self.scheduler.prep(self)
#
# 	def __repr__(self):
# 		base = super().__repr__()
# 		if self.scheduler is not None:
# 			title, *rest = base.split('\n')
# 			sch = repr(self.scheduler)
# 			base = '\n'.join((title, sch, *rest))
# 		return base
#
# 	def load_state_dict(self, state_dict, strict=True):
# 		if self.scheduler is not None and 'scheduler' in state_dict:
# 			self.scheduler.load_state_dict(state_dict['scheduler'])
# 		super().load_state_dict(state_dict, strict=strict)
#
# 	def state_dict(self, *args, **kwargs):
# 		state_dict = super().state_dict(*args, **kwargs)
# 		if self.scheduler is not None:
# 			state_dict['scheduler'] = self.scheduler.state_dict()
# 		return state_dict


# region Complex extensions


class Complex_Optimizer(OptimizerBase):
	def __init__(self, **optims):
		self.__dict__['optims'] = None
		super().__init__(defaults=None)
		self.__dict__['optims'] = optims
		self.group_params()
		# self._update_groups()
		# self._update_named_groups()
		# self.__dict__['param_groups'] = [grp for grp in chain(*[o.param_groups for o in optims.values()])]

	def group_params(self):
		self.param_groups = sum([[p for p in o.param_groups] for o in self.optims.values()], [])

	def _update_groups(self):
		self.param_groups = self._param_group_gen()

	def _param_group_gen(self):
		for optim in self.optims.values():
			for grp in optim.param_groups:
				yield grp
		self._update_groups()

	def _update_named_groups(self):
		self.named_param_groups = self._named_param_group_gen()

	def _named_param_group_gen(self):
		for key, optim in self.optims.items():
			for grp in optim.param_groups:
				yield (key, grp)
		self._update_named_groups()

	def add_param_group(self, *args, **kwargs): # probably shouldnt be used
		return
		# raise Exception('invalid for complex optimizers')

	def load_state_dict(self, state_dict):
		for name, optim in self.optims.items():
			optim.load_state_dict(state_dict[name])
		self.group_params() # important to get loaded param groups

	def state_dict(self):
		return {name:optim.state_dict() for name, optim in self.optims.items()}

	def zero_grad(self):
		for optim in self.optims.values():
			optim.zero_grad()

	def step(self, closure=None): # can and perhaps should be overridden
		for optim in self.optims.values():
			optim.step(closure)

	def __iter__(self):
		return iter(self.optims)
	def items(self):
		return self.optims.items()

	def __len__(self):
		return len(self.optims)

	def __getitem__(self, item):
		return self.optims[item]
	def __setitem__(self, key, value):
		# raise NotImplementedError
		self.optims[key] = value
	def __delitem__(self, key):
		del self.optims[key]

	def __getattr__(self, item):
		try:
			return super().__getattribute__(item)
		except AttributeError:
			return self.optims[item]
	def __setattr__(self, key, value):
		# raise NotImplementedError
		if isinstance(value, OptimizerBase):
			# raise NotImplementedError
			self.__setitem__(key, value)
		else:
			super().__setattr__(key, value)
	def __delattr__(self, item):
		try:
			super().__delattr__(item)
		except AttributeError:
			self.__delitem__(item)

	def __str__(self):
		s = ['Complex-Optimizer(']

		for k,v in self.optims.items():
			s.append('  {}: {}'.format(k,str(v).replace('\n', '\n    ')))

		s.append(')')

		return '\n'.join(s)


# endregion

## region old


# def get_optimizer(optim_type, parameters, lr=1e-3, weight_decay=0, momentum=0, beta1=.9, beta2=.999, **optim_args):
# 	if optim_type == 'sgd':
# 		optimizer = O.SGD(parameters,lr=lr, weight_decay=weight_decay, momentum=momentum, **optim_args)
# 	elif optim_type == 'rmsprop':
# 		optimizer = O.RMSprop(parameters, lr=lr, weight_decay=weight_decay, momentum=momentum, **optim_args)
# 	elif optim_type == 'adam':
# 		optimizer = O.Adam(parameters, lr=lr, weight_decay=weight_decay, betas=(beta1, beta2), **optim_args)
# 	elif optim_type == 'adamw':
# 		optimizer = O.AdamW(parameters, lr=lr, weight_decay=weight_decay, betas=(beta1, beta2), **optim_args)
# 	elif optim_type == 'ranger':
# 		optimizer = Ranger(parameters, lr=lr, weight_decay=weight_decay,betas=(beta1, beta2), **optim_args) #alpha=0.5, k=6, N_sma_threshhold=5, betas=(.95,0.999))
# 	elif optim_type == 'cg':
# 		optimizer = Conjugate_Gradient(parameters, **optim_args)
# 	elif optim_type == 'rprop':
# 		optimizer = O.Rprop(parameters, lr=lr)
# 	elif optim_type == 'adagrad':
# 		optimizer = O.Adagrad(parameters, lr=lr, weight_decay=weight_decay)
# 	elif optim_type == 'adadelta':
# 		optimizer = O.Adadelta(parameters, lr=lr, weight_decay=weight_decay)
# 	else:
# 		assert False, "Unknown optimizer type: " + optim_type
# 	return optimizer

# def default_create_optim(parameters, info): # info is a train.Config
#
# 	kwargs = {
#
# 		'optim_type': info.pull('optim_type'),
#
# 		'parameters': parameters,
#
# 		'lr': info.pull('lr', 1e-3),
# 		'weight_decay': info.pull('weight_decay', 0),
# 		'momentum': info.pull('momentum', 0),
#
# 		'beta1': info.pull('beta1', .9),
# 		'beta2': info.pull('beta2', .999),
# 	}
#
# 	return get_optimizer(**kwargs)
#
# def default_create_scheduler(optimizer, info):
#
# 	if 'scheduler_type' not in info:
# 		return None
#
# 	name = info.pull('scheduler_type')
#
# 	freq = info.pull('scheduler_freq', None)
# 	# if freq is None:
# 	# 	print('Scheduler will step at the end of each epoch')
# 	# else:
# 	# 	print(f'Scheduler will step every {freq} iterations')
# 	cut_after = info.pull('scheduler_cut_after', None)
#
# 	min_lr = info.pull('scheduler_min_lr', 0.)
# 	last = info.pull('scheduler_last_epoch', -1)
#
# 	common = {
# 		'cut_after': cut_after,
# 		'freq': freq,
# 	}
#
# 	if name == 'step':
#
# 		factor = info.pull('scheduler_decay', 0.1)
#
# 		step_size = info.pull('scheduler_step')
# 		out = StepLR(optimizer, step_size=step_size, gamma=factor, last_epoch=last,
# 		             req_loss=False, **common)
# 	elif name == 'plateau':
#
# 		factor = info.pull('scheduler_decay', 0.1)
# 		patience = info.pull('scheduler_patience', 10)
# 		cooldown = info.pull('scheduler_cooldown', 0)
#
# 		out = ReduceOnPlateau(optimizer, factor=factor, patience=patience, verbose=True,
# 		                                       min_lr=min_lr, cooldown=cooldown,
# 		                      req_loss=True, **common)
#
# 	elif name == 'lambda':
#
# 		func = info.pull('scheduler_lambda')
#
# 		out = LambdaLR(optimizer, lr_lambda=func, last_epoch=last, **common)
#
# 	elif name == 'cos':
# 		num_steps = info.pull('scheduler_total_steps', None) # num_steps
# 		if num_steps is None:
# 			if freq is None or freq <= 0:
# 				raise Exception('cos scheduler needs to know the max number of steps')
# 			num_steps = info.pull('scheduler_total_iterations', '<>training.step_limit') // freq \
# 				- info.pull('scheduler_early_stop', 0)
#
# 		eta_min = min_lr
#
# 		out = CosineAnnealing(optimizer, T_max=num_steps, eta_min=eta_min, last_epoch=last,
# 		                      req_loss=False, **common)
#
# 	else:
# 		raise Exception(f'unknown name {name}')
#
# 	return out

# region old

# class Base_Scheduler(O.lr_scheduler._LRScheduler):
#
# 	def __init__(self, *args, cut_after=None, freq=None, req_loss=None, **kwargs):
# 		if freq is not None and freq <= 0:
# 			freq = None
# 		self.freq = freq
# 		self.cut_after = cut_after
# 		self.req_loss = req_loss
# 		super().__init__(*args, **kwargs)
#
# 	def requires_loss(self):
# 		return self.req_loss
#
# 	def epoch_end(self, *args, **kwargs):
# 		if self.freq is None:
# 			self.step(*args, **kwargs)
#
# 	def maintain(self, step, *args, **kwargs):
# 		if self.freq is not None and step % self.freq == 0:
# 			self.step(*args, **kwargs)
#
# 	def step(self, *args, **kwargs):
# 		if self.cut_after is not None and self._step_count > self.cut_after:
# 			return
# 		return super().step(*args, **kwargs)
#
# class StepLR(Base_Scheduler, O.lr_scheduler.StepLR):
# 	def __str__(self):
# 		return self.__repr__()
#
# 	def __repr__(self):
# 		return 'StepLR(step={}, gamma={})'.format(self.step_size, self.gamma)
#
# class LambdaLR(Base_Scheduler, O.lr_scheduler.LambdaLR):
#
# 	def __str__(self):
# 		return self.__repr__()
#
# 	def __repr__(self):
# 		return f'LambdaLR(lambda={self.lr_lambdas[0]})'
#
# class MultiStepLR(Base_Scheduler, O.lr_scheduler.MultiStepLR):
# 	def __str__(self):
# 		return self.__repr__()
#
# 	def __repr__(self):
# 		return 'MultiStepLR(milestones={}, gamma={})'.format(self.milestones, self.gamma)
#
# class ReduceOnPlateau(Base_Scheduler, O.lr_scheduler.ReduceLROnPlateau):
# 	def __str__(self):
# 		return self.__repr__()
#
# 	def __repr__(self):
# 		return 'ReduceOnPlateau(factor={}, patience={}, cooldown={})'.format(self.factor, self.patience, self.cooldown)
#
# class CosineAnnealing(Base_Scheduler, O.lr_scheduler.CosineAnnealingLR):
# 	def __str__(self):
# 		return self.__repr__()
#
# 	def __repr__(self):
# 		return 'CosineAnnealing(T_max={},eta_min={})'.format(self.T_max, self.eta_min)

# endregion


# class Conjugate_Gradient(Optimizer):
# 	def __init__(self, params, step_size, nsteps, residual_tol=1e-10, ret_stats=False):
# 		super().__init__(params, {'step-size': step_size,
#                                   'nsteps': nsteps,
#                                   'res-tol': residual_tol,
#                                   'ret_stats': ret_stats})
#
# 		self.stats = [StatsMeter('v-norm', 'n-norm', 'alpha') for _ in self.param_groups]
#
# 	def cg_solve(self, apply_A, b, x0=None, res_tol=1e-10, nsteps=10):
#
# 		if x0 is None:
# 			x = torch.zeros(*b.size())
# 			r = b.clone()
# 		else:
# 			x = x0
# 			r = b - apply_A(x)
#
# 		p = r.clone()
# 		rdotr = r @ r
# 		for i in range(nsteps):
# 			Ap = apply_A(p)
# 			alpha = rdotr / (p @ Ap)
# 			x += alpha * p
# 			r -= alpha * Ap
# 			new_rdotr = r @ r
# 			beta = new_rdotr / rdotr
# 			p = r + beta * p
# 			rdotr = new_rdotr
# 			if rdotr < res_tol:
# 				break
#
# 		return x
#
# 	def step(self, apply_A):
#
# 		for group, stats in zip(self.param_groups, self.stats):
#
# 			# get v (from backward pass)
# 			p = nn.parameters_to_vector(group)
# 			v = p.grad
#
# 			# compute n with cg
# 			n = self.cg_solve(apply_A, v, res_tol=group['res-tol'], nsteps=group['nsteps'])
#
# 			# compute learning rate
# 			alpha = (2 * group['step-size'] / (v @ n)).sqrt()
#
# 			p.data.add( n.mul_(alpha) )
# 			nn.vector_to_parameters(p, group)
#
# 			if group['ret_stats']:
# 				stats.update('alpha', alpha)
# 				stats.update('v-norm', v.norm())
# 				stats.update('n-norm', n.norm())
				
				
# endregion


